import json
import logging
from typing import Dict, List, Tuple, Union, Mapping, Optional, Sequence, cast

import betterproto

from qm.program import Program
from qm.jobs.qm_job import QmJob
from qm.octave import QmOctaveConfig
from qm.persistence import BaseStore
from qm.octave.qm_octave import QmOctave
from qm.api.models.devices import Polarity
from qm.api.frontend_api import FrontendApi
from qm.jobs.pending_job import QmPendingJob
from qm.jobs.job_queue_old_api import QmQueue
from qm.jobs.simulated_job import SimulatedJob
from qm.api.simulation_api import SimulationApi
from qm.jobs.running_qm_job import RunningQmJob
from qm.utils.config_utils import get_fem_config
from qm.octave.octave_manager import OctaveManager
from qm.simulate.interface import SimulationConfig
from qm.elements_db import ElementsDB, init_elements
from qm.utils.types_utils import convert_object_type
from qm.program.ConfigBuilder import convert_msg_to_config
from qm.elements.up_converted_input import UpconvertedInput
from qm._QmJobErrors import InvalidDigitalInputPolarityError
from qm.api.v2.job_api.job_api import JobApiWithDeprecations
from qm.octave._calibration_names import COMMON_OCTAVE_PREFIX
from qm.api.job_manager_api import create_job_manager_from_api
from qm.api.models.capabilities import OPX_FEM_IDX, ServerCapabilities
from qm.jobs.job_queue_with_deprecations import QmQueueWithDeprecations
from qm.api.models.compiler import CompilerOptionArguments, standardize_compiler_params
from qm.type_hinting.config_types import StandardPort, DictQuaConfig, PortReferenceType
from qm.elements.element_inputs import MixInputs, SingleInput, static_set_mixer_correction
from qm.type_hinting.general import Value, Number, PathLike, NumpySupportedFloat, NumpySupportedValue
from qm.octave.octave_mixer_calibration import AutoCalibrationParams, OctaveMixerCalibration, MixerCalibrationResults
from qm.grpc.frontend import (
    JobExecutionStatus,
    JobExecutionStatusLoading,
    JobExecutionStatusPending,
    JobExecutionStatusRunning,
    JobExecutionStatusCompleted,
)
from qm.grpc.qua_config import (
    QuaConfig,
    QuaConfigQuaConfigV1,
    QuaConfigPortReference,
    QuaConfigMicrowaveFemDec,
    QuaConfigDigitalInputPortDec,
    QuaConfigDigitalInputPortDecPolarity,
)
from qm.exceptions import (
    QmValueError,
    JobCancelledError,
    ErrorJobStateError,
    FunctionInputError,
    InvalidConfigError,
    AnotherJobIsRunning,
    CantCalibrateElementError,
    UnsupportedCapabilityError,
)

logger = logging.getLogger(__name__)


class QuantumMachine:
    def __init__(
        self,
        machine_id: str,
        pb_config: QuaConfig,
        frontend_api: FrontendApi,
        capabilities: ServerCapabilities,
        store: BaseStore,
        octave_manager: OctaveManager,
        octave_config: Optional[QmOctaveConfig] = None,
    ):
        self._id = machine_id
        self._config = pb_config
        self._frontend = frontend_api

        self._capabilities = capabilities

        self._simulation_api = SimulationApi(self._frontend.connection_details)
        self._job_manager = create_job_manager_from_api(frontend_api)

        self._store = store
        self._queue = QmQueue(
            config=self._config,
            quantum_machine_id=self._id,
            frontend_api=self._frontend,
            capabilities=self._capabilities,
            store=self._store,
        )

        self._elements: ElementsDB = init_elements(
            pb_config, frontend_api, machine_id=machine_id, octave_config=octave_config
        )
        self._octave = QmOctave(self, octave_manager)

    @property
    def _octave_calibration_elements(self) -> ElementsDB:
        return ElementsDB({k: v for k, v in self._elements.items() if k.startswith(COMMON_OCTAVE_PREFIX)})

    @property
    def id(self) -> str:
        return self._id

    @property
    def queue(self) -> Union[QmQueue, QmQueueWithDeprecations]:
        return self._queue

    @property
    def octave(self) -> QmOctave:
        return self._octave

    def close(self) -> bool:
        """Closes the quantum machine.

        Returns:
            `True` if the close request succeeded, raises an exception
            otherwise.
        """
        return self._frontend.close_quantum_machine(self._id)

    def simulate(
        self,
        program: Program,
        simulate: SimulationConfig,
        compiler_options: Optional[CompilerOptionArguments] = None,
        *,
        strict: Optional[bool] = None,
        flags: Optional[List[str]] = None,
    ) -> SimulatedJob:
        """Simulates the outputs of a deterministic QUA program.

        Equivalent to ``execute()`` with ``simulate=SimulationConfig`` (see example).

        Note:
            A simulated job does not support calling QuantumMachine API functions.

        Args:
            program: A QUA ``program()`` object to execute
            simulate: If given, will be simulated instead of executed.
            compiler_options: Optional arguments for compilation.
            strict: This parameter is deprecated, please use `compiler_options`
            flags: This parameter is deprecated, please use `compiler_options`
        Returns:
            A ``QmJob`` object (see QM Job API).
        """
        standardized_compiler_options = standardize_compiler_params(compiler_options, strict, flags)
        job = cast(
            SimulatedJob, self.execute(program, simulate=simulate, compiler_options=standardized_compiler_options)
        )
        return job

    def execute(
        self,
        program: Program,
        duration_limit: int = 1000,
        data_limit: int = 20000,
        force_execution: int = False,
        dry_run: int = False,
        simulate: Optional[SimulationConfig] = None,
        compiler_options: Optional[CompilerOptionArguments] = None,
        *,
        strict: Optional[bool] = None,
        flags: Optional[List[str]] = None,
    ) -> RunningQmJob:
        """Executes a program and returns a job object to keep track of execution and get
        results.

        Note:
            Calling execute will halt any currently running program and clear the current
            queue. If you want to add a job to the queue, use qm.queue.add()

        Args:
            program: A QUA ``program()`` object to execute
            duration_limit: This parameter is ignored and will be removed in future versions
            data_limit: This parameter is ignored and will be removed in future versions
            force_execution: This parameter is ignored and will be removed in future versions
            dry_run: This parameter is ignored and will be removed in future versions
            simulate: If given, will be simulated instead of executed.
            compiler_options: Optional arguments for compilation.
            strict: This parameter is deprecated, please use `compiler_options`
            flags: This parameter is deprecated, please use `compiler_options`
        Returns:
            A ``QmJob`` object (see QM Job API).
        """
        if not isinstance(program, Program):
            raise Exception("program argument must be of type qm.program.Program")
        if program.metadata.uses_command_timestamps and not self._capabilities.supports_command_timestamps:
            raise UnsupportedCapabilityError("timestamping commands is supported from QOP 2.2 or above")
        if program.metadata.uses_fast_frame_rotation and not self._capabilities.supports_fast_frame_rotation:
            raise UnsupportedCapabilityError("fast frame rotation is supported from QOP 2.2 or above")

        standardized_compiler_options = standardize_compiler_params(compiler_options, strict, flags)

        if simulate is not None:
            job_id, simulated_response_part = self._simulation_api.simulate(
                self._get_pb_config(), program, simulate, standardized_compiler_options
            )
            return SimulatedJob(
                job_id=job_id,
                frontend_api=self._frontend,
                capabilities=self._capabilities,
                store=self._store,
                simulated_response=simulated_response_part,
            )

        self._queue.clear()
        current_running_job = self.get_running_job()
        if current_running_job is not None:
            current_running_job.cancel()
        pending_job = self._queue.add(program, standardized_compiler_options)
        logger.info("Executing program")
        return pending_job.wait_for_execution(timeout=5)

    def compile(
        self,
        program: Program,
        compiler_options: Optional[CompilerOptionArguments] = None,
    ) -> str:
        """Compiles a QUA program to be executed later. The returned `program_id`
        can then be directly added to the queue. For a detailed explanation
        see [Precompile Jobs](../Guides/features.md#precompile-jobs).

        Args:
            program: A QUA program
            compiler_options: Optional arguments for compilation

        Returns:
            a program_id str

        Example:
            ```python
            program_id = qm.compile(program)
            pending_job = qm.queue.add_compiled(program_id)
            job = pending_job.wait_for_execution()
            ```
        """
        if program.metadata.uses_command_timestamps and not self._capabilities.supports_command_timestamps:
            raise UnsupportedCapabilityError("timestamping commands is supported from QOP 2.2 or above")

        if program.metadata.uses_fast_frame_rotation and not self._capabilities.supports_fast_frame_rotation:
            raise UnsupportedCapabilityError("fast frame rotation is supported from QOP 2.2 or above")

        logger.info("Compiling program")

        if compiler_options is None:
            compiler_options = CompilerOptionArguments()
        return self._frontend.compile(self._id, program.qua_program, compiler_options)

    def list_controllers(self) -> Tuple[str, ...]:
        """Gets a list with the defined controllers in this qm

        Returns:
            The names of the controllers configured in this qm
        """
        config = self._get_config_as_object()
        return tuple(config.control_devices) or tuple(config.controllers)

    def set_mixer_correction(
        self,
        mixer: str,
        intermediate_frequency: Number,
        lo_frequency: Number,
        values: Tuple[float, float, float, float],
    ) -> None:
        """Sets the correction matrix for correcting gain and phase imbalances
        of an IQ mixer for the supplied intermediate frequency and LO frequency.

        Args:
            mixer (str): the name of the mixer, as defined in the
                configuration
            intermediate_frequency (Union[int|float]): the intermediate
                frequency for which to apply the correction matrix
            lo_frequency (int): the LO frequency for which to apply the
                correction matrix
            values (tuple):

                tuple is of the form (v00, v01, v10, v11) where
                the matrix is
                | v00 v01 |
                | v10 v11 |

        Note:

            Currently, the OPX does not support multiple mixer calibration entries.
            This function will accept IF & LO frequencies written in the config file,
            and will update the correction matrix for all the elements with the given
            mixer/frequencies combination when the program started.

            It’s not recommended to use this method while a job is running.
            To change the calibration values for a running job,
            use job.set_element_correction
        """
        static_set_mixer_correction(
            self._frontend,
            self._id,
            mixer,
            intermediate_frequency=intermediate_frequency,
            lo_frequency=lo_frequency,
            values=values,
        )

    def calibrate_element(
        self,
        qe: str,
        lo_if_dict: Optional[Mapping[float, Tuple[float, ...]]] = None,
        save_to_db: bool = True,
        params: Optional[AutoCalibrationParams] = None,
    ) -> MixerCalibrationResults:
        """Calibrate the up converters associated with a given element for the given LO & IF frequencies.

        - Frequencies can be given as a dictionary with LO frequency as the key and a list of IF frequencies for every LO
        - If no frequencies are given calibration will occur according to LO & IF declared in the element
        - The function need to be run for each element separately
        - The results are saved to a database for later use

        Args:
            qe (str): The name of the element for calibration
            lo_if_dict ([Mapping[float, Tuple[float, ...]]]): a dictionary with LO frequency as the key and
                a list of IF frequencies for every LO
            save_to_db (bool): If true (default), The calibration
                parameters will be saved to the calibration database
            params: Optional calibration parameters
        """
        inst = self._elements[qe]
        if not isinstance(inst.input, UpconvertedInput):
            raise CantCalibrateElementError(
                f"Element {qe} has input of type {type(inst.input)} and is not connected to an Octave. "
                f"Hence, it cannot be calibrated."
            )
        if params is None:
            params = AutoCalibrationParams()

        if self.get_running_job():
            raise AnotherJobIsRunning

        if lo_if_dict is None:
            lo_if_dict = {inst.input.lo_frequency: (inst.intermediate_frequency,)}
        client = self._octave._octave_manager._get_client_from_port(inst.input.port)
        res = OctaveMixerCalibration(quantum_machine=self, client=client).calibrate(
            element=inst,
            lo_if_dict=lo_if_dict,
            params=params,
        )

        if save_to_db:
            calibration_db = self._octave._octave_manager._octave_config._calibration_db
            if calibration_db is None:
                logger.warning("No calibration db found, can't save results")
            else:
                calibration_db.update_calibration_result(res, inst.input.port, "auto")

        key = (inst.input.lo_frequency, cast(float, inst.input.gain))
        if key in res:
            qe_cal = res[key]
            if inst.intermediate_frequency in qe_cal.image:
                inst.input.set_output_dc_offset(i_offset=qe_cal.i0, q_offset=qe_cal.q0)

        for (lo_freq, _), lo_cal in res.items():
            for if_freq, if_cal in lo_cal.image.items():
                fine_cal = if_cal.fine
                self.set_mixer_correction(
                    mixer=inst.input.mixer,
                    intermediate_frequency=if_freq,
                    lo_frequency=lo_freq,
                    values=fine_cal.correction,
                )
        return res

    def set_intermediate_frequency(self, element: str, freq: float) -> None:
        """Sets the intermediate frequency of the element

        Args:
            element (str): the name of the element whose intermediate
                frequency will be updated
            freq (float): the intermediate frequency to set to the given
                element
        """
        element_inst = self._elements[element]
        element_inst.set_intermediate_frequency(freq)

    def get_output_dc_offset_by_element(self, element: str, iq_input: str) -> float:
        """Get the current DC offset of the OPX analog output channel associated with an element.

        Args:
            element: the name of the element to get the correction for
            iq_input: the port name as appears in the element config.
                Options:

                `'single'`
                    for an element with a single input

                `'I'` or `'Q'`
                    for an element with mixer inputs

        Returns:
            the offset, in volts
        """
        config = self._get_pb_config()
        input_instance = self._elements[element].input
        if isinstance(input_instance, SingleInput):
            port = input_instance.port
        elif isinstance(input_instance, MixInputs):
            if iq_input == "I":
                port = input_instance.i_port
            elif iq_input == "Q":
                port = input_instance.q_port
            else:
                raise ValueError(f"Port must be I or Q, got {iq_input}.")
        else:
            raise ValueError(f"Element {element} of type {type(input_instance)} does not have a 'port' property.")

        port_number = port.number

        specific_fem = get_fem_config(config, port)
        if isinstance(specific_fem, QuaConfigMicrowaveFemDec):
            raise InvalidConfigError(f"Port num {port_number} is a MW-FEM port, has no offset")
        if port_number in specific_fem.analog_outputs:
            return specific_fem.analog_outputs[port_number].offset
        else:
            raise InvalidConfigError(f"Port num {port_number} does not exist")

    def set_output_dc_offset_by_element(
        self,
        element: str,
        input: Union[str, Tuple[str, str], List[str]],
        offset: Union[float, Tuple[float, float], List[float]],
    ) -> None:
        """Set the current DC offset of the OPX analog output channel associated with an element.

        Args:
            element (str): the name of the element to update the
                correction for
            input (Union[str, Tuple[str,str], List[str]]): the input
                name as appears in the element config. Options:

                `'single'`
                    for an element with a single input

                `'I'` or `'Q'` or a tuple ('I', 'Q')
                    for an element with mixer inputs
            offset (Union[float, Tuple[float,float], List[float]]): The
                dc value to set to, in volts. Ranges
                from -0.5 to 0.5 - 2^-16 in steps of 2^-16.

        Examples:
            ```python
            qm.set_output_dc_offset_by_element('flux', 'single', 0.1)
            qm.set_output_dc_offset_by_element('qubit', 'I', -0.01)
            qm.set_output_dc_offset_by_element('qubit', ('I', 'Q'), (-0.01, 0.05))
            ```

        Note:

            If the sum of the DC offset and the largest waveform data-point exceed the range,
            DAC output overflow will occur and the output will be corrupted.
        """
        element_inst = self._elements[element]
        if isinstance(element_inst.input, MixInputs):
            if isinstance(input, (list, tuple)):
                if not set(input) <= {"I", "Q"}:
                    raise FunctionInputError(f"Input names should be 'I' or 'Q', got {input}")
                if not (isinstance(offset, (list, tuple)) and len(input) == len(offset)):
                    raise FunctionInputError(
                        f"input should be two iterables of the same size," f"got input = {input} and offset = {offset}"
                    )
                kwargs = {f"{k.lower()}_offset": v for k, v in zip(input, offset)}
            elif isinstance(input, str):
                if input not in {"I", "Q"}:
                    raise FunctionInputError(f"Input names should be 'I' or 'Q', got {input}")
                if not isinstance(offset, (int, float)):
                    raise FunctionInputError(f"Input should be int or float, got {type(offset)}")
                kwargs = {f"{input.lower()}_offset": offset}
            else:
                raise ValueError(f"Invalid input - {input}")
            element_inst.input.set_output_dc_offset(**kwargs)
        elif isinstance(element_inst.input, SingleInput):
            if input != "single":
                raise ValueError(f"Invalid input - {input} while the element has a single input")
            if not isinstance(offset, (int, float)):
                raise FunctionInputError(f"Input should be int or float, got {type(offset)}")
            element_inst.input.set_output_dc_offset(offset)
        else:
            raise ValueError(
                f"Element {element} with input of type {element_inst.input.__class__.__name__} "
                f"does not support dc offset setting."
            )

    def set_output_filter_by_element(
        self,
        element: str,
        input: str,
        feedforward: Sequence[NumpySupportedFloat],
        feedback: Sequence[NumpySupportedFloat],
    ) -> None:
        """Sets the intermediate frequency of the element

        Args:
            element: the name of the element whose ports filters will be
                updated
            input: the input name as appears in the element config.
                Options:

                `'single'`
                    for an element with single input

                `'I'` or `'Q'`
                    for an element with mixer inputs
            feedforward: the values for the feedforward filter
            feedback: the values for the feedback filter

        """
        logger.debug(
            f"Setting output filter of port '{input}' on element '{element}' "
            + f"to (feedforward: {feedforward}, feedback: {feedback})"
        )
        input_inst = self._elements[element].input
        if isinstance(input_inst, MixInputs):
            input_inst.set_output_filter(input, feedforward, feedback)
        elif isinstance(input_inst, SingleInput):
            input_inst.set_output_filter(feedforward, feedback)
        else:
            raise AttributeError(
                f"Element {element} of type {input_inst.__class__.__name__} " f"does not support filter setting."
            )

    def set_input_dc_offset_by_element(self, element: str, output: str, offset: float) -> None:
        """Set the current DC offset of the OPX analog input channel associated with an element.

        Args:
            element (str): the name of the element to update the
                correction for
            output (str): the output key name as appears in the element
                config under 'outputs'.
            offset (float): the dc value to set to, in volts. Ranges from -0.5 to 0.5 - 2^-16 in steps of
                2^-16.

        Note:
            If the sum of the DC offset and the largest waveform data-point exceed the range,
            DAC output overflow will occur and the output will be corrupted.
        """
        element_instance = self._elements[element]
        element_instance.set_input_dc_offset(output, offset)

    def get_input_dc_offset_by_element(self, element: str, output: str) -> float:
        """Get the current DC offset of the OPX analog input channel associated with an element.

        Args:
            element: the name of the element to get the correction for
            output: the output key name as appears in the element config
                under 'outputs'.

        Returns:
            The offset, in volts
        """
        config = self._get_pb_config()

        element_obj = self._elements[element]
        outputs = element_obj._config.outputs
        if output in outputs:
            port = outputs[output]
        else:
            raise Exception("Output does not exist")

        specific_fem = get_fem_config(config, port)
        if isinstance(specific_fem, QuaConfigMicrowaveFemDec):
            raise InvalidConfigError(f"Port num {port.number} is a MW-FEM port, has no offset")

        if port.number in specific_fem.analog_inputs:
            return specific_fem.analog_inputs[port.number].offset
        else:
            raise InvalidConfigError(f"Port num {port.number} does not exist")

    def get_digital_delay(self, element: str, digital_input: str) -> int:
        """Gets the delay of the digital input of the element

        Args:
            element: The name of the element to get the delay for
            digital_input: The digital input name as appears in the
                element's config

        Returns:
            The delay
        """
        element_instance = self._elements[element]
        return element_instance.get_output_digital_delay(digital_input)

    def set_digital_delay(self, element: str, digital_input: str, delay: int) -> None:
        """Sets the delay of the digital input of the element

        Args:
            element (str): The name of the element to update delay for
            digital_input (str): The digital input name as appears in
                the element's config
            delay (int): The delay value to set to, in ns.
        """
        element_instance = self._elements[element]
        element_instance.set_output_digital_delay(digital_input, delay)

    def get_digital_buffer(self, element: str, digital_input: str) -> int:
        """Gets the buffer for digital input of the element

        Args:
            element (str): The name of the element to get the buffer for
            digital_input (str): The digital input name as appears in
                the element's config

        Returns:
            The buffer
        """
        element_instance = self._elements[element]
        return element_instance.get_output_digital_buffer(digital_input_name=digital_input)

    def set_digital_buffer(self, element: str, digital_input: str, buffer: int) -> None:
        """Sets the buffer for digital input of the element

        Args:
            element (str): The name of the element to update buffer for
            digital_input (str): the digital input name as appears in
                the element's config
            buffer (int): The buffer value to set to, in ns.
        """
        element_instance = self._elements[element]
        element_instance.set_output_digital_buffer(digital_input, buffer)

    def get_time_of_flight(self, element: str) -> int:
        """Gets the *time of flight*, associated with a measurement element.

        This is the amount of time between the beginning of a measurement pulse applied to element
        and the time that the data is available to the controller for demodulation or streaming.

        Args:
            element (str): The name of the element to get time of flight
                for

        Returns:
            The time of flight, in ns
        """
        element_object = self._elements[element]
        return element_object.time_of_flight

    def get_smearing(self, element: str) -> int:
        """Gets the *smearing* associated with a measurement element.

        This is a broadening of the raw results acquisition window, to account for dispersive broadening
        in the measurement elements (readout resonators etc.) The acquisition window will be broadened
        by this amount on both sides.

        Args:
            element (str): The name of the element to get smearing for

        Returns:
            The smearing, in ns.
        """
        element_object = self._elements[element]
        return element_object.smearing

    @property
    def io1(self) -> Dict[str, Value]:
        return self.get_io1_value()

    @io1.setter
    def io1(self, value: Value) -> None:
        self.set_io1_value(value)

    @property
    def io2(self) -> Dict[str, Value]:
        return self.get_io1_value()

    @io2.setter
    def io2(self, value: Value) -> None:
        self.set_io2_value(value)

    def set_io1_value(self, value_1: Value) -> None:
        """Sets the value of ``IO1``.

        This can be used later inside a QUA program as a QUA variable ``IO1`` without declaration.
        The type of QUA variable is inferred from the python type passed to ``value_1``,
        according to the following rule:

        int -> int
        float -> fixed
        bool -> bool

        Args:
            value_1 (Union[float,bool,int]): the value to be placed in
                ``IO1``
        """
        self.set_io_values(value_1=value_1)

    def set_io2_value(self, value_2: Value) -> None:
        """Sets the value of ``IO1``.

        This can be used later inside a QUA program as a QUA variable ``IO2`` without declaration.
        The type of QUA variable is inferred from the python type passed to ``value_2``,
        according to the following rule:

        int -> int
        float -> fixed
        bool -> bool

        Args:
            value_2 (Union[float, bool, int]): the value to be placed in
                ``IO1``
        """
        self.set_io_values(value_2=value_2)

    def set_io_values(
        self,
        value_1: Optional[NumpySupportedValue] = None,
        value_2: Optional[NumpySupportedValue] = None,
    ) -> None:
        """Sets the values of ``IO1`` and ``IO2``

        This can be used later inside a QUA program as a QUA variable ``IO1``, ``IO2`` without declaration.
        The type of QUA variable is inferred from the python type passed to ``value_1``, ``value_2``,
        according to the following rule:

        int -> int
        float -> fixed
        bool -> bool

        Args:
            value_1 (Optional[Union[float, bool, int]]): the value to be
                placed in ``IO1``
            value_2 (Optional[Union[float, bool, int]]): the value to be
                placed in ``IO2``
        """

        if value_1 is None and value_2 is None:
            return

        try:
            if value_1 is not None:
                logger.debug(f"Setting value '{value_1}' to IO1")
                value_1 = convert_object_type(value_1)

            if value_2 is not None:
                logger.debug(f"Setting value '{value_2}' to IO2")
                value_2 = convert_object_type(value_2)
        except QmValueError as e:
            raise QmValueError(f"Failed to set_io_values: {e.message}") from e

        self._frontend.set_io_values(self._id, [value_1, value_2])

    def get_io1_value(self) -> Dict[str, Value]:
        """Gets the data stored in ``IO1``

        No inference is made on type.

        Returns:
            A dictionary with data stored in ``IO1``. (Data is in all
            three format: ``int``, ``float`` and ``bool``)
        """
        return self.get_io_values()[0]

    def get_io2_value(self) -> Dict[str, Value]:
        """Gets the data stored in ``IO2``

        No inference is made on type.

        Returns:
            A dictionary with data from the second IO register. (Data is
            in all three format: ``int``, ``float`` and ``bool``)
        """
        return self.get_io_values()[1]

    def get_io_values(self) -> List[Dict[str, Value]]:
        """Gets the data stored in both ``IO1`` and ``IO2``

        No inference is made on type.

        Returns:
            A list that contains dictionaries with data from the IO
            registers. (Data is in all three format: ``int``, ``float``
            and ``bool``)
        """
        resp1, resp2 = self._frontend.get_io_values(self._id)

        return [
            {
                "io_number": 1,
                "int_value": resp1.int_value,
                "fixed_value": resp1.double_value,
                "boolean_value": resp1.boolean_value,
            },
            {
                "io_number": 2,
                "int_value": resp2.int_value,
                "fixed_value": resp2.double_value,
                "boolean_value": resp2.boolean_value,
            },
        ]

    def _get_pb_config(self) -> QuaConfig:
        return self._frontend.get_quantum_machine_config(self._id)

    def _get_config_as_object(self) -> QuaConfigQuaConfigV1:
        return self._get_pb_config().v1_beta

    def get_config(self) -> DictQuaConfig:
        """Gets the current config of the qm

        Returns:
            A dictionary with the QMs config
        """
        config = self._get_pb_config()
        self._config = config
        return convert_msg_to_config(config)

    def save_config_to_file(self, filename: PathLike) -> None:
        """Saves the qm current config to a file

        Args:
            filename: The name of the file where the config will be saved
        """
        json_str = json.dumps(self.get_config())
        with open(filename, "w") as writer:
            writer.write(json_str)

    def get_running_job(self) -> Optional[Union[QmJob, JobApiWithDeprecations]]:
        """Gets the currently running job. Returns None if there isn't one."""

        job_id = self._job_manager.get_running_job(self._id)
        if job_id is None:
            return None

        try:
            pending_job = QmPendingJob(
                job_id=job_id,
                machine_id=self._id,
                frontend_api=self._frontend,
                capabilities=self._capabilities,
                store=self._store,
            )
            return pending_job.wait_for_execution(timeout=10.0)
        except JobCancelledError:
            # In case that the job has finished between the GetRunningJon and the
            # wait for execution
            return None

    def get_job(self, job_id: str) -> Union[QmJob, QmPendingJob, JobApiWithDeprecations]:
        """
        Get a job based on the job_id.

        Args:
            job_id: A list of jobs ids
        Returns:
            The job
        """
        status: JobExecutionStatus = self._job_manager.get_job_execution_status(job_id, self._id)
        _, status_inst = betterproto.which_one_of(status, "status")
        if isinstance(status_inst, (JobExecutionStatusRunning, JobExecutionStatusCompleted)):
            return QmJob(
                job_id=job_id,
                machine_id=self._id,
                frontend_api=self._frontend,
                capabilities=self._capabilities,
                store=self._store,
            )
        if isinstance(status_inst, (JobExecutionStatusPending, JobExecutionStatusLoading)):
            return QmPendingJob(
                job_id=job_id,
                machine_id=self._id,
                frontend_api=self._frontend,
                capabilities=self._capabilities,
                store=self._store,
            )

        raise ErrorJobStateError(
            f"job {self._id} encountered an error",
            error_list=[value.string_value for value in status.error.error_messages.values],
        )

    def set_digital_input_threshold(self, port: PortReferenceType, threshold: float) -> None:
        """Sets the threshold of the digital input

        Args:
            port: The port, in tuple form, to be updated
            threshold: The threshold in volts
        """
        controller_name, fem_number, port_number = _standardize_port(port)
        self._frontend.set_digital_input_threshold(self._id, controller_name, fem_number, port_number, threshold)

    def _get_digital_input_port(self, port: PortReferenceType) -> QuaConfigDigitalInputPortDec:
        config = self._get_pb_config()
        controller_name, fem_number, port_number = _standardize_port(port)
        port_obj = QuaConfigPortReference(controller=controller_name, fem=fem_number, number=port_number)
        fem_config = get_fem_config(config, port_obj)
        if port_obj.number in fem_config.digital_inputs:
            return fem_config.digital_inputs[port_number]
        raise InvalidConfigError("Digital input port not found")

    def get_digital_input_threshold(self, port: PortReferenceType) -> float:
        """Gets the threshold of the digital input

        Args:
            port: The port, in tuple form, to be updated

        Returns:
            The threshold in volts
        """
        return self._get_digital_input_port(port).threshold

    def set_digital_input_deadtime(self, port: PortReferenceType, deadtime: int) -> None:
        """Sets the threshold of the digital input

        Args:
            port: The port, in tuple form, to be updated
            deadtime: The deadtime in ns
        """
        controller_name, fem_number, port_number = _standardize_port(port)
        self._frontend.set_digital_input_dead_time(self._id, controller_name, fem_number, port_number, deadtime)

    def get_digital_input_deadtime(self, port: PortReferenceType) -> int:
        """Gets the deadtime of the digital input

        Args:
            port: The port, in tuple form, to be updated

        Returns:
            The deadtime in ns
        """
        return self._get_digital_input_port(port).deadtime

    def set_digital_input_polarity(self, port: PortReferenceType, polarity: str) -> None:
        """Sets the polarity of the digital input

        Args:
            port: The port, in tuple form, to be updated
            polarity: The polarity: "RISING" or "FALLING"
        """
        try:
            polarity_enum = Polarity[polarity]
        except KeyError:
            raise InvalidDigitalInputPolarityError(
                f"Invalid value for polarity {polarity}. Valid values are: 'RISING' or 'FALLING'."
            )

        controller_name, fem_number, port_number = _standardize_port(port)
        self._frontend.set_digital_input_polarity(self._id, controller_name, fem_number, port_number, polarity_enum)

    def get_digital_input_polarity(self, port: PortReferenceType) -> str:
        """Gets the polarity of the digital input

        Args:
            port: The port, in tuple form, to be updated

        Returns:
            The polarity
        """
        return QuaConfigDigitalInputPortDecPolarity(self._get_digital_input_port(port).polarity).name


def _standardize_port(port: PortReferenceType) -> StandardPort:
    if len(port) == 2:
        return port[0], OPX_FEM_IDX, port[1]
    return cast(StandardPort, port)



import logging
from dataclasses import dataclass
from typing import Dict, List, Tuple, Union, Optional, cast

import numpy as np

from qm.api.v2.job_api import JobApi
from qm.jobs.running_qm_job import RunningQmJob
from qm.octave.calibration_utils import convert_to_correction
from qm.octave._calibration_names import SavedVariablesNames as Names

Array = np.typing.NDArray[np.cdouble]


logger = logging.getLogger(__name__)


def _polyfit2d(x: Array, y: Array, z: Array, kx: int = 2, ky: int = 2, order: Optional[int] = None) -> Array:
    """
    Two-dimensional polynomial fitting by least squares.
    Fits the functional form f(x,y) = z.

    Notes
    -----
    Resultant fit can be plotted with:
    np.polynomial.polynomial.polygrid2d(x, y, soln.reshape((kx+1, ky+1)))

    Parameters
    ----------
    x, y: array-like, 1d
        x and y coordinates.
        len(x) mest be z.shape[0] and len(y) must be z.shape[1]
    z: np.ndarray, 2d
        Surface to fit.
    kx, ky: int, default is 3
        Polynomial order in x and y, respectively.
    order: int or None, default is None
        If None, all coefficients up to maximum kx, ky, i.e. up to and including x^kx*y^ky, are considered.
        If int, coefficients up to a maximum of kx+ky <= order are considered.

    Returns
    -------
    Return parameters from np.linalg.lstsq.

    soln: np.ndarray
        Array of polynomial coefficients.
    residuals: np.ndarray
    rank: int
    s: np.ndarray

    """

    # grid coords
    x, y = np.meshgrid(x, y)
    # coefficient array, up to x^kx, y^ky
    coefficients = np.ones((kx + 1, ky + 1))

    # solve array
    a = np.zeros((coefficients.size, x.size))

    # for each coefficient produce array x^i, y^j
    for index, (j, i) in enumerate(np.ndindex(*coefficients.shape)):
        # do not include powers greater than order
        if order is not None and i + j > order:
            arr = np.zeros_like(x)
        else:
            arr = coefficients[i, j] * x**i * y**j
        a[index] = arr.ravel()

    # do least-squares fitting
    return cast(Array, np.linalg.lstsq(a.T, np.ravel(z), rcond=None)[0])


@dataclass
class FitResult:
    """
    Result of a two-dimensional polynomial fitting.
    """

    pol: Array
    x_min: float
    y_min: float
    theta: float
    pol_: Array


def _paraboloid2d_fit(x: Array, y: Array, z: Array) -> FitResult:
    """
    Two-dimensional paraboloid fitting by least squares.
    Fits the functional form a0 + a1*x + a2*y + a3*x**2 + a4*x*y + a5*y**2 = z.

    Parameters
    ----------
    x, y: array-like, 1d
        x and y coordinates.
        len(x) mest be z.shape[0] and len(y) must be z.shape[1]
    z: np.ndarray, 2d
        Surface to fit.

    Returns
    -------
    A dictionary with the fitted polynom and the axis of the minimal point:

    pol
    x_min
    y_min

    """

    p = _polyfit2d(x, y, z, order=2).take([0, 1, 3, 2, 4, 6])

    theta = np.arctan2(-p[4], p[3] - p[5]) / 2.0
    c = np.cos(theta)
    s = np.sin(theta)

    p_ = (
        np.array(
            [
                [1, 0, 0, 0, 0, 0],  # 1
                [0, c, -s, 0, 0, 0],  # x
                [0, s, c, 0, 0, 0],  # y
                [0, 0, 0, c**2, -c * s, s**2],  # x**2
                [0, 0, 0, 2 * c * s, c**2 - s**2, -2 * c * s],  # x*y
                [0, 0, 0, s**2, c * s, c**2],  # y**2
            ]
        )
        @ p
    )

    x0_ = -p_[1] / (2 * p_[3])
    y0_ = -p_[2] / (2 * p_[5])

    xy_min = np.array([[c, s], [-s, c]]) @ [x0_, y0_]

    return FitResult(pol=p, x_min=xy_min[0], y_min=xy_min[1], theta=theta, pol_=p_)


def _get_reshaped_data(
    job: Union[RunningQmJob, JobApi], n_samples: int, offset: int, count: int, *names: str
) -> Dict[str, Array]:
    data_first = offset * n_samples**2
    data_last = (offset + count) * n_samples**2

    name_to_result_handle = {}
    for name in names:
        result_handle = job.result_handles.get(name)
        assert result_handle is not None
        name_to_result_handle[name] = result_handle
        result_handle.wait_for_values(data_last)

    raw_data = {}
    for name in names:
        result_handle = name_to_result_handle[name]
        raw_data[name] = result_handle.fetch(slice(data_first, data_last), flat_struct=True)

    full_data = {}
    for name, value in raw_data.items():
        full_data[name] = np.array(value).reshape((-1, n_samples, n_samples))

    return full_data


@dataclass
class CorrectionsDebugData:
    """
    Debug data for the lo calibration correction.
    """

    dc_gain: float
    dc_phase: float
    dc_correction: Tuple[float, float, float, float]


@dataclass
class LOAnalysisDebugData:
    """
    Debug data for the LO calibration analysis.
    """

    i_scan: Array
    q_scan: Array
    lo: Array
    fit: FitResult
    corrections: CorrectionsDebugData


def _get_and_analyze_lo_data(
    job: Union[RunningQmJob, JobApi], lo_res: int, offset: int, count: int
) -> Tuple[Array, Array, List[LOAnalysisDebugData]]:
    data = _get_reshaped_data(job, lo_res, offset, count, Names.i_scan, Names.q_scan, Names.lo)

    calibration_pulse_amp = 0.25  # TODO(nofek): take it from the right place

    i_scan_full = data[Names.i_scan] * calibration_pulse_amp
    q_scan_full = data[Names.q_scan] * calibration_pulse_amp
    lo_full = data[Names.lo]

    q0_shift = []
    i0_shift = []
    debug_data = []
    for ind in range(i_scan_full.shape[0]):
        i_scan = i_scan_full[ind]
        q_scan = q_scan_full[ind]
        lo = lo_full[ind]

        q = q_scan[0, :]
        i = i_scan[:, 0]
        if np.min(lo) < 0:
            logger.warning("POSSIBLE OVERFLOW!!!")  # TODO: deal with it
        ii, iq = divmod(int(np.argmin(lo)), lo_res)
        ll = min(iq, lo_res - 1 - iq, ii, lo_res - 1 - ii, 7)
        if ll > 3:
            fit = _paraboloid2d_fit(
                q[iq - ll : iq + ll + 1], i[ii - ll : ii + ll + 1], lo[ii - ll : ii + ll + 1, iq - ll : iq + ll + 1]
            )
        else:
            fit = _paraboloid2d_fit(q, i, lo)

        q0_shift.append(fit.x_min)
        i0_shift.append(fit.y_min)

        _I2c, _IQc, _Q2c = fit.pol.take([5, 4, 3])
        _n = np.sqrt(_I2c * _Q2c)
        _I2c, _IQc, _Q2c = _I2c / _n, _IQc / _n, _Q2c / _n

        dc_gain = (_Q2c / _I2c) ** 0.25
        dc_phase = np.arcsin(-_IQc / 2) / 2

        c = np.cos(dc_phase) / np.cos(2 * dc_phase)
        s = np.sin(dc_phase) / np.cos(2 * dc_phase)
        g = dc_gain

        dc_correction = (g * c, g * s, s / g, c / g)

        # Matching the meaning of the phase and gain to that of the calibration
        dc_phase = np.sin(dc_phase) / np.cos(2 * dc_phase)
        dc_gain = (1 + 2 * (dc_gain - 1)) ** 0.5 - 1

        corrections = CorrectionsDebugData(dc_gain=dc_gain, dc_phase=dc_phase, dc_correction=dc_correction)

        debug_data.append(LOAnalysisDebugData(i_scan=i_scan, q_scan=q_scan, lo=lo, fit=fit, corrections=corrections))

    return np.array(i0_shift), np.array(q0_shift), debug_data


@dataclass
class ImageDataAnalysisResult:
    """
    Debug data for the image calibration analysis.
    """

    phase: float
    gain: float
    correction: Tuple[float, float, float, float]
    g_scan: Array
    p_scan: Array
    fit: FitResult
    image: Array


def _get_and_analyze_image_data(
    job: Union[RunningQmJob, JobApi], image_res: int, offset: int, count: int
) -> List[ImageDataAnalysisResult]:
    data = _get_reshaped_data(job, image_res, offset, count, Names.g_scan, Names.p_scan, Names.image)

    g_scan_full = data[Names.g_scan]
    p_scan_full = data[Names.p_scan]
    image_full = data[Names.image]

    results = []
    for ind in range(g_scan_full.shape[0]):
        g_scan = g_scan_full[ind]
        p_scan = p_scan_full[ind]
        image = image_full[ind]
        p = p_scan[0, :]
        g = g_scan[:, 0]

        if np.min(image) < 0:
            logger.warning("POSSIBLE OVERFLOW!!!")  # TODO deal with it

        ig, ip = divmod(int(np.argmin(image)), image_res)
        ll = min(ip, image_res - 1 - ip, ig, image_res - 1 - ig, 5)

        sg = g[ig - ll : ig + ll + 1]
        sp = p[ip - ll : ip + ll + 1]

        s_im = image[ig - ll : ig + ll + 1, ip - ll : ip + ll + 1]

        fit = _paraboloid2d_fit(sp, sg, s_im)

        results.append(
            ImageDataAnalysisResult(
                phase=fit.x_min,
                gain=fit.y_min,
                correction=convert_to_correction(fit.y_min, fit.x_min),
                g_scan=g_scan,
                p_scan=p_scan,
                fit=fit,
                image=image,
            )
        )

    return results



import sys
from logging import LogRecord
from logging.handlers import BufferingHandler

from datadog_api_client.v2.api.logs_api import LogsApi
from datadog_api_client import ApiClient, Configuration
from datadog_api_client.v2.model.http_log import HTTPLog
from datadog_api_client.v2.model.http_log_item import HTTPLogItem
from datadog_api_client.v2.model.content_encoding import ContentEncoding

from qm.version import __version__


class DatadogHandler(BufferingHandler):
    def __init__(self, user_id: str, organization: str, user_token: str, session_id: str):
        super().__init__(capacity=100)
        self._user_id = user_id
        self._organization = organization
        self._user_token = user_token
        self._session_id = session_id

    def _prepare_record(self, record: LogRecord) -> HTTPLogItem:
        return HTTPLogItem(
            ddsource="python",
            hostname=self._user_id,
            message=record.msg,
            service=record.name,
            status=record.levelname,
            session_id=str(self._session_id),
            python_version=sys.version,
            qua_version=__version__,
        )

    @property
    def logs_as_http(self) -> HTTPLog:
        _logs = []
        for record in self.buffer:
            _logs.append(self._prepare_record(record))
        return HTTPLog(_logs)  # type: ignore[no-untyped-call]

    def send_logs(self) -> None:
        configuration = Configuration(  # type: ignore[no-untyped-call]
            api_key={"apiKeyAuth": self._user_token},
            server_variables={"site": "datadoghq.eu"},
        )
        with ApiClient(configuration) as api_client:
            api_instance = LogsApi(api_client)  # type: ignore[no-untyped-call]
            response = api_instance.submit_log(content_encoding=ContentEncoding.GZIP, body=self.logs_as_http)
            print(response)

    def flush(self) -> None:
        self.send_logs()
        super().flush()



import dataclasses
from collections import deque
from typing import Deque, Union

from multidict import MultiDict

RECEIVED_HEADERS_MAX_SIZE = 10000


@dataclasses.dataclass
class DebugData:
    received_headers: Deque[MultiDict[Union[str, bytes]]] = dataclasses.field(
        default_factory=lambda: deque(maxlen=RECEIVED_HEADERS_MAX_SIZE)
    )

    def append(self, received_metadata: MultiDict[Union[str, bytes]]) -> None:
        self.received_headers.append(received_metadata)



import logging
from typing import Type, Tuple, Union, AsyncIterator

from qm.simulate import interface
from qm.api.base_api import BaseApi
from qm.grpc.qua_config import QuaConfig
from qm.utils.async_utils import run_async
from qm.program import Program, load_config
from qm.utils.protobuf_utils import LOG_LEVEL_MAP
from qm.simulate.interface import SimulationConfig
from qm.type_hinting.config_types import DictQuaConfig
from qm.api.models.server_details import ConnectionDetails
from qm.exceptions import QMSimulationError, FailedToExecuteJobException
from qm.api.models.compiler import CompilerOptionArguments, get_request_compiler_options
from qm.grpc.results_analyser import SimulatorSamplesResponse, PullSimulatorSamplesRequest
from qm.grpc.frontend import (
    FrontendStub,
    DensityMatrix,
    InterOpxAddress,
    InterOpxChannel,
    SimulationRequest,
    InterOpxConnection,
    SimulatedResponsePart,
    ExecutionRequestSimulate,
    GetSimulatedQuantumStateRequest,
    InterOpxConnectionAddressToAddress,
    InterOpxConnectionChannelToChannel,
)

logger = logging.getLogger(__name__)


def create_simulation_request(
    config: Union[DictQuaConfig, QuaConfig],
    program: Program,
    simulate: SimulationConfig,
    compiler_options: CompilerOptionArguments,
) -> SimulationRequest:
    if not isinstance(program, Program):
        raise Exception("program argument must be of type qm.program.Program")

    request = SimulationRequest()
    msg_config = config if isinstance(config, QuaConfig) else load_config(config)
    msg_config.v1_beta._unknown_fields = b""  # The above correction is for a return of an unknown fields from
    # an old GW, that changes the flow of the config parsing (in the GW, and to a buggy one).
    # Check a year from now (Feb/2024) if this is still necessary.
    request.config = msg_config

    if isinstance(simulate, SimulationConfig):
        request.simulate = ExecutionRequestSimulate(
            duration=simulate.duration,
            include_analog_waveforms=simulate.include_analog_waveforms,
            include_digital_waveforms=simulate.include_digital_waveforms,
            extra_processing_timeout_ms=simulate.extraProcessingTimeoutInMs,
        )
        request = simulate.update_simulate_request(request)

        for connection in simulate.controller_connections:
            if not isinstance(connection.source, type(connection.target)):
                raise Exception(
                    f"Unsupported InterOpx connection. Source is "
                    f"{type(connection.source).__name__} but target is "
                    f"{type(connection.target).__name__}"
                )

            if isinstance(connection.source, interface.InterOpxAddress) and isinstance(
                connection.target, interface.InterOpxAddress
            ):
                con = InterOpxConnection(
                    address_to_address=InterOpxConnectionAddressToAddress(
                        source=InterOpxAddress(
                            controller=connection.source.controller,
                            left=connection.source.is_left_connection,
                        ),
                        target=InterOpxAddress(
                            controller=connection.target.controller,
                            left=connection.target.is_left_connection,
                        ),
                    )
                )
            elif isinstance(connection.source, interface.InterOpxChannel) and isinstance(
                connection.target, interface.InterOpxChannel
            ):
                con = InterOpxConnection(
                    channel_to_channel=InterOpxConnectionChannelToChannel(
                        source=InterOpxChannel(
                            controller=connection.source.controller,
                            channel_number=connection.source.channel_number,
                        ),
                        target=InterOpxChannel(
                            controller=connection.target.controller,
                            channel_number=connection.target.channel_number,
                        ),
                    )
                )
            else:
                raise Exception(
                    f"Unsupported InterOpx connection. Source is "
                    f"{type(connection.source).__name__}. Supported types are "
                    f"InterOpxAddress "
                    f"or InterOpxChannel"
                )

            request.controller_connections.append(con)

    request.high_level_program = program.qua_program
    request.high_level_program.compiler_options = get_request_compiler_options(compiler_options)
    return request


class SimulationApi(BaseApi[FrontendStub]):
    def __init__(self, connection_details: ConnectionDetails):
        super().__init__(connection_details)
        self._timeout = None

    @property
    def _stub_class(self) -> Type[FrontendStub]:
        return FrontendStub

    def simulate(
        self,
        config: Union[DictQuaConfig, QuaConfig],
        program: Program,
        simulate: SimulationConfig,
        compiler_options: CompilerOptionArguments,
    ) -> Tuple[str, SimulatedResponsePart]:
        if type(program) is not Program:
            raise Exception("program argument must be of type qm.program.Program")

        request = create_simulation_request(config, program, simulate, compiler_options)
        logger.info("Simulating program")

        response = run_async(self._stub.simulate(request, timeout=self._timeout))

        messages = [(LOG_LEVEL_MAP[msg.level], msg.message) for msg in response.messages]

        config_messages = [(LOG_LEVEL_MAP[msg.level], msg.message) for msg in response.config_validation_errors]

        job_id = response.job_id

        for lvl, msg in messages:
            logger.log(lvl, msg)

        for lvl, msg in config_messages:
            logger.log(lvl, msg)

        if not response.success:
            logger.error("Job " + job_id + " failed. Failed to execute program.")
            for error in response.simulated.errors:
                logger.error(f"Simulation error: {error}")
            raise FailedToExecuteJobException(job_id)

        return job_id, response.simulated

    def get_simulated_quantum_state(self, job_id: str) -> DensityMatrix:
        request = GetSimulatedQuantumStateRequest(job_id=job_id)
        response = run_async(self._stub.get_simulated_quantum_state(request, timeout=self._timeout))

        if response.ok:
            return response.state

        raise QMSimulationError("Error while pulling quantum state")

    def pull_simulator_samples(
        self, job_id: str, include_analog: bool, include_digital: bool
    ) -> AsyncIterator[SimulatorSamplesResponse]:
        request = PullSimulatorSamplesRequest(
            job_id=job_id,
            include_analog=include_analog,
            include_digital=include_digital,
            include_all_connections=True,  # TODO: Check whether it should appear
        )
        return self._stub.pull_simulator_samples(request)



from typing import Dict, Optional

from qm.api.v2.job_api.generic_apis import ElementGenericApi
from qm.api.v2.job_api.element_input_api import ElementInputApi
from qm.grpc.v2 import GetIntermediateFrequencyRequest, SetIntermediateFrequencyRequest
from qm.api.v2.job_api.element_port_api import MwOutputApi, AnalogOutputApi, DigitalInputApi


class PortNotFound(KeyError):
    def __init__(self, key: str):
        self._key = key

    def __str__(self) -> str:
        return f"Port {self._key} was not found."


class NoMicrowaveOutputError(KeyError):
    def __str__(self) -> str:
        return "Microwave output is not set"


class ElementApi(ElementGenericApi):
    def set_intermediate_frequency(self, frequency: float) -> None:
        request = SetIntermediateFrequencyRequest(job_id=self._id, qe=self._element_id, frequency=frequency)
        self._run(self._stub.set_intermediate_frequency(request, timeout=self._timeout))

    def get_intermediate_frequency(self) -> float:
        request = GetIntermediateFrequencyRequest(job_id=self._id, qe=self._element_id)
        response = self._run(self._stub.get_intermediate_frequency(request, timeout=self._timeout))
        return response.frequency


class DigitalInputs(Dict[str, DigitalInputApi]):
    def __missing__(self, key: str) -> None:
        raise PortNotFound(key)


class AnalogOutputs(Dict[str, AnalogOutputApi]):
    def __missing__(self, key: str) -> None:
        raise PortNotFound(key)


class JobElement:
    def __init__(
        self,
        api: ElementApi,
        input_api: ElementInputApi,
        outputs_apis: Dict[str, AnalogOutputApi],
        digital_inputs_apis: Dict[str, DigitalInputApi],
        microwave_output_api: Optional[MwOutputApi],
    ):
        self._api = api
        self._input_api = input_api
        self._outputs_apis = AnalogOutputs(outputs_apis)
        self._digital_inputs_apis = DigitalInputs(digital_inputs_apis)
        self._microwave_output_api = microwave_output_api

    def set_intermediate_frequency(self, frequency: float) -> None:
        self._api.set_intermediate_frequency(frequency)

    def get_intermediate_frequency(self) -> float:
        return self._api.get_intermediate_frequency()

    @property
    def input(self) -> ElementInputApi:
        return self._input_api

    @property
    def outputs(self) -> AnalogOutputs:
        return self._outputs_apis

    @property
    def digital_inputs(self) -> DigitalInputs:
        return self._digital_inputs_apis

    @property
    def microwave_output(self) -> MwOutputApi:
        if self._microwave_output_api is None:
            raise NoMicrowaveOutputError
        return self._microwave_output_api

    @property
    def has_mw_output(self) -> bool:
        return self._microwave_output_api is not None



from typing import List, Type, Tuple, Union, TypeVar

import numpy as np
import numpy.typing as npt

from qm.grpc import qua

MessageExpressionType = qua.QuaProgramAnyScalarExpression
MessageArrayVarType = qua.QuaProgramArrayVarRefExpression
MessageVarType = qua.QuaProgramVarRefExpression
MessageVariableType = Union[MessageArrayVarType, MessageVarType]
MessageVariableOrExpression = Union[MessageExpressionType, MessageVariableType]


ArrayT = TypeVar("ArrayT")
PyArrayType = Union[List[ArrayT], npt.NDArray[ArrayT], Tuple[ArrayT, ...]]

NpTypes = Union[np.floating, np.integer, np.bool_, np.longdouble]

PyIntType = Union[int, np.integer]
PyFloatType = Union[float, np.floating, np.longdouble]
PyNumberType = Union[PyFloatType, PyIntType]
PyNumberArrayType = PyArrayType[PyNumberType]
PyBoolType = Union[bool, np.bool_]

AllPyTypes = Union[PyBoolType, PyNumberType]

ForEachValuesType = Union[PyNumberArrayType, Tuple[PyNumberArrayType, ...]]

T = TypeVar("T")
OneOrMore = Union[T, PyArrayType[T]]


class fixed:
    pass


VariableDeclarationType = Union[Type[int], Type[bool], Type[float], Type[fixed]]



import logging
from typing import Tuple, Union, Generic, TypeVar, Optional, Sequence

import numpy
from dependency_injector.wiring import Provide, inject

from qm.api.frontend_api import FrontendApi
from qm.grpc.general_messages import Matrix
from qm.api.models.capabilities import ServerCapabilities
from qm.api.models.devices import MixerInfo, AnalogOutputPortFilter
from qm.containers.capabilities_container import CapabilitiesContainer
from qm.type_hinting.general import Number, NumpySupportedFloat, NumpySupportedNumber
from qm.grpc.qua_config import (
    QuaConfigMixInputs,
    QuaConfigSingleInput,
    QuaConfigMultipleInputs,
    QuaConfigDacPortReference,
    QuaConfigSingleInputCollection,
    QuaConfigMicrowaveInputPortReference,
)

logger = logging.getLogger(__name__)


def _set_single_output_port_dc_offset(
    frontend_api: FrontendApi,
    machine_id: str,
    element_name: str,
    input_name: str,
    offset: NumpySupportedNumber,
) -> None:
    offset = float(offset)
    logger.debug(f"Setting DC offset of input '{input_name}' on element '{element_name}' to '{offset}'")
    frontend_api.set_output_dc_offset(machine_id, element_name, input_name, offset)


def _create_taps_filter(
    feedforward: Sequence[NumpySupportedFloat], feedback: Sequence[NumpySupportedFloat]
) -> AnalogOutputPortFilter:
    for name, instance in zip(["feedforward", "feedback"], [feedforward, feedback]):
        if not isinstance(instance, (numpy.ndarray, list)):
            raise TypeError(f"{name} must be a list, or a numpy array. Got {type(instance)}.")
    return AnalogOutputPortFilter(feedforward=[float(x) for x in feedforward], feedback=[float(x) for x in feedback])


@inject
def static_set_mixer_correction(
    frontend_api: FrontendApi,
    machine_id: str,
    mixer: str,
    intermediate_frequency: Union[int, float],
    lo_frequency: Union[int, float],
    values: Tuple[float, float, float, float],
    capabilities: ServerCapabilities = Provide[CapabilitiesContainer.capabilities],
) -> None:
    # TODO - this function is here (and not under MixedInputsElement) to support backwards the direct calling to mixer
    #  Once it is changed, one can put this function under the element

    if not isinstance(values, (tuple, list)) or len(values) != 4:
        raise Exception("correction values must have 4 items")

    float_values = [float(x) for x in values]
    if not all((-2 < x <= (2 - 2 ** (-16))) for x in float_values):
        logger.warning(
            "At least one of the correction values are out of range. "
            f"values should be between -2 and 2 - 2 ** (-16), got {float_values}. "
            f"Not setting the correction matrix."
        )
        return
    correction_matrix = Matrix(*float_values)

    mixer_lo_frequency_double = 0.0
    mixer_intermediate_frequency_double = 0.0
    if capabilities.supports_double_frequency:
        mixer_lo_frequency_double = float(lo_frequency)
        mixer_intermediate_frequency_double = abs(float(intermediate_frequency))

    mixer_info = MixerInfo(
        mixer=mixer,
        frequency_negative=intermediate_frequency < 0,
        lo_frequency=int(lo_frequency),
        intermediate_frequency=abs(int(intermediate_frequency)),
        lo_frequency_double=mixer_lo_frequency_double,
        intermediate_frequency_double=mixer_intermediate_frequency_double,
    )
    frontend_api.set_correction(machine_id, mixer_info, correction_matrix)


ElementInputGRPCType = TypeVar(
    "ElementInputGRPCType",
    QuaConfigSingleInput,
    QuaConfigMixInputs,
    QuaConfigSingleInputCollection,
    QuaConfigMultipleInputs,
    QuaConfigMicrowaveInputPortReference,
    None,
)


class ElementInput(Generic[ElementInputGRPCType]):
    def __init__(self, name: str, config: ElementInputGRPCType, frontend_api: FrontendApi, machine_id: str):
        self._name = name
        self._config: ElementInputGRPCType = config
        self._frontend = frontend_api
        self._id = machine_id


class NoInput(ElementInput[None]):
    pass


class MicrowaveInput(ElementInput[QuaConfigMicrowaveInputPortReference]):
    pass


class SingleInput(ElementInput[QuaConfigSingleInput]):
    @property
    def port(self) -> QuaConfigDacPortReference:
        return self._config.port

    def set_output_dc_offset(self, offset: float) -> None:
        _set_single_output_port_dc_offset(self._frontend, self._id, self._name, "single", offset)

    def set_output_filter(
        self,
        feedforward: Sequence[NumpySupportedFloat],
        feedback: Sequence[NumpySupportedFloat],
    ) -> None:
        analog_filter = _create_taps_filter(feedforward, feedback)
        self._frontend.set_output_filter_taps(self._id, self._name, "single", analog_filter)


class MultipleInputs(ElementInput[QuaConfigMultipleInputs]):
    pass


class SingleInputCollection(ElementInput[QuaConfigSingleInputCollection]):
    pass


class MixInputs(ElementInput[QuaConfigMixInputs]):
    @property
    def i_port(self) -> QuaConfigDacPortReference:
        return self._config.i

    @property
    def q_port(self) -> QuaConfigDacPortReference:
        return self._config.q

    @property
    @inject
    def lo_frequency(self, capabilities: ServerCapabilities = Provide[CapabilitiesContainer.capabilities]) -> float:
        if capabilities.supports_double_frequency:
            return self._config.lo_frequency_double
        return self._config.lo_frequency

    def set_lo_frequency(self, value: float) -> None:
        self._set_config_lo_frequency(value)

    @inject
    def _set_config_lo_frequency(
        self, value: float, capabilities: ServerCapabilities = Provide[CapabilitiesContainer.capabilities]
    ) -> None:
        freq = float(value)
        logger.debug(f"Setting element '{self._name}' LO frequency to '{freq}'.")
        self._config.lo_frequency = int(freq)
        self._config.lo_frequency_double = 0.0
        if capabilities.supports_double_frequency:
            self._config.lo_frequency_double = float(freq)

    @property
    def mixer(self) -> str:
        return self._config.mixer

    def set_output_dc_offset(self, i_offset: Optional[float] = None, q_offset: Optional[float] = None) -> None:
        if i_offset is not None:
            _set_single_output_port_dc_offset(self._frontend, self._id, self._name, "I", i_offset)
        if q_offset is not None:
            _set_single_output_port_dc_offset(self._frontend, self._id, self._name, "Q", q_offset)

    def set_output_filter(
        self,
        input_name: str,
        feedforward: Sequence[NumpySupportedFloat],
        feedback: Sequence[NumpySupportedFloat],
    ) -> None:
        analog_filter = _create_taps_filter(feedforward, feedback)
        self._frontend.set_output_filter_taps(
            self._id,
            self._name,
            input_name,
            analog_filter,
        )

    def set_mixer_correction(
        self,
        intermediate_frequency: Number,
        lo_frequency: Number,
        values: Tuple[float, float, float, float],
    ) -> None:
        static_set_mixer_correction(
            self._frontend,
            self._id,
            mixer=self.mixer,
            intermediate_frequency=intermediate_frequency,
            lo_frequency=lo_frequency,
            values=values,
        )



import warnings
from enum import Enum
from typing import Tuple

from qm.jobs.base_job import QmBaseJob
from qm.utils import deprecation_message
from qm.grpc.general_messages import Matrix
from qm.results import StreamingResultFetcher
from qm.api.job_result_api import JobResultServiceApi
from qm._report import ExecutionError, ExecutionReport


class AcquiringStatus(Enum):
    AcquireStopped = 0
    NoDataToAcquire = 1
    HasDataToAcquire = 2


class RunningQmJob(QmBaseJob):
    @property
    def manager(self) -> None:
        """
        The QM object where this job lives
        """
        warnings.warn(
            deprecation_message(
                method="RunningQmJob.manager",
                deprecated_in="1.1.0",
                removed_in="1.2.0",
                details="QMJob no longer has 'manager' property",
            ),
            DeprecationWarning,
            stacklevel=2,
        )
        return None

    @property
    def result_handles(self) -> StreamingResultFetcher:
        """

        Returns:
            The handles that this job generated
        """
        return StreamingResultFetcher(
            self._id,
            JobResultServiceApi(self._frontend.connection_details, self._id),
            self._store,
            self._capabilities,
        )

    def cancel(self) -> bool:
        """
        Cancels the job
        Sames as `job.halt()`

        Returns:
             `True` if the job was canceled successfully
        """
        return self.halt()

    def halt(self) -> bool:
        """Halts the job on the opx

        Returns:
             `True` if the job was canceled successfully"""
        return self._job_manager.halt(self._id)

    def resume(self) -> bool:
        """Resumes a program that was halted using the [pause][qm.qua._dsl.pause] statement"""
        return self._job_manager.resume(self._id)

    def is_paused(self) -> bool:
        """Returns:
        Returns `True` if the job was paused from QUA.
        """
        return self._job_manager.is_paused(self._id)

    def _is_job_running(self) -> bool:
        """Returns:
        Returns `True` if the job is running
        """
        return self._job_manager.is_job_running(self._id)

    def _is_data_acquiring(self) -> AcquiringStatus:
        """Returns the data acquiring status.
        The possible statuses are: AcquireStopped, NoDataToAcquire,  HasDataToAcquire

        Returns:
            An AcquiringStatus enum object
        """
        status = self._job_manager.is_data_acquiring(self._id)
        return AcquiringStatus(status.value)

    def execution_report(self) -> ExecutionReport:
        """Get runtime errors report for this job. See [Runtime errors](../Guides/error.md#runtime-errors).

        Returns:
            An object holding the errors that this job generated.
        """
        errors = [
            ExecutionError.create_from_grpc_message(item)
            for item in JobResultServiceApi(self._frontend.connection_details, self.id).get_job_errors()
        ]
        return ExecutionReport(self._id, errors)

    def set_element_correction(
        self, element: str, correction: Tuple[float, float, float, float]
    ) -> Tuple[float, float, float, float]:
        r"""Sets the correction matrix for correcting gain and phase imbalances
        of an IQ mixer associated with an element.

        Changes will only be done to the current job!

        Values will be rounded to an accuracy of $2^{-16}$.
        Valid values for the correction values are between $-2$ and $(2 - 2^{-16})$.

        Warning - the correction matrix can increase the output voltage which might result in an
        overflow.

        Args:
            element (str): the name of the element to update the
                correction for
            correction (tuple):

                tuple is of the form (v00, v01, v10, v11) where
                the matrix is
                $\begin{pmatrix} v_{00} & v_{01} \\ v_{10} & v_{11}\end{pmatrix}$

        Returns:
            The correction matrix, after rounding to the OPX resolution.
        """
        correction_matrix = Matrix(v00=correction[0], v01=correction[1], v10=correction[2], v11=correction[3])

        return self._job_manager.set_element_correction(self._id, element, correction_matrix)

    def get_element_correction(self, element: str) -> Tuple[float, float, float, float]:
        """Gets the correction matrix for correcting gain and phase imbalances
        of an IQ mixer associated with an element.

        Args:
            element (str): The name of the element to update the
                correction for

        Returns:
            The current correction matrix
        """
        return self._job_manager.get_element_correction(self._id, element)



import logging
import warnings
from typing_extensions import Mapping  # Mapping in 3.8 does not support class indexing
from typing import Dict, List, Tuple, Union, TypeVar, KeysView, Optional, Generator, ItemsView, ValuesView

from qm.persistence import BaseStore
from qm.utils import deprecation_message
from qm.api.v2.job_result_api import JobResultApi
from qm.api.job_result_api import JobResultServiceApi
from qm.api.models.capabilities import ServerCapabilities
from qm.utils.general_utils import run_until_with_timeout
from qm.StreamMetadata import StreamMetadata, StreamMetadataError
from qm.results.single_streaming_result_fetcher import SingleStreamingResultFetcher
from qm.results.multiple_streaming_result_fetcher import MultipleStreamingResultFetcher
from qm.results.base_streaming_result_fetcher import (
    JobResultSchema,
    JobResultItemSchema,
    BaseStreamingResultFetcher,
    _parse_dtype,
)

logger = logging.getLogger(__name__)


_T = TypeVar("_T")


class StreamingResultFetcher(Mapping[str, Optional[BaseStreamingResultFetcher]]):
    """Access to the results of a QmJob

    This object is created by calling [QmJob.result_handles][qm.jobs.running_qm_job.RunningQmJob.result_handles]

    Assuming you have an instance of StreamingResultFetcher:
    ```python
        job_results: StreamingResultFetcher
    ```
    This object is iterable:

    ```python
        for name, handle in job_results:
            print(name)
    ```

    Can detect if a name exists:

    ```python
    if "somename" in job_results:
        print("somename exists!")
        handle = job_results.get("somename")
    ```
    """

    def __init__(
        self,
        job_id: str,
        service: Union[JobResultServiceApi, JobResultApi],
        store: BaseStore,
        capabilities: ServerCapabilities,
    ) -> None:
        self._job_id = job_id
        self._service = service
        self._store = store
        self._schema = StreamingResultFetcher._load_schema(self._service)
        self._capabilities = capabilities

        self._all_results = self._get_job_results()

    def _get_job_results(self) -> Dict[str, BaseStreamingResultFetcher]:
        _all_results = {}
        stream_metadata_errors, stream_metadata_dict = self._get_stream_metadata()
        for name, item_schema in self._schema.items.items():
            stream_metadata = stream_metadata_dict.get(name)
            result: BaseStreamingResultFetcher
            if item_schema.is_single:
                result = SingleStreamingResultFetcher(
                    schema=item_schema,
                    service=self._service,
                    store=self._store,
                    stream_metadata_errors=stream_metadata_errors,
                    stream_metadata=stream_metadata,
                    capabilities=self._capabilities,
                )
            else:
                result = MultipleStreamingResultFetcher(
                    job_results=self,
                    schema=item_schema,
                    service=self._service,
                    store=self._store,
                    stream_metadata_errors=stream_metadata_errors,
                    stream_metadata=stream_metadata,
                    capabilities=self._capabilities,
                )
            _all_results[name] = result
        return _all_results

    def __len__(self) -> int:
        return len(self._all_results)

    def __getitem__(self, item: str) -> Optional[BaseStreamingResultFetcher]:
        return self.get(item)

    def __getattr__(self, item: str) -> Optional[BaseStreamingResultFetcher]:
        if item == "shape" or item == "__len__":
            return (
                None  # this is here because of a bug in pycharm debugger: ver: 2022.3.2 build #PY-223.8617.48 (24/1/23)
            )
        return self.get(item)

    def _get_stream_metadata(self) -> Tuple[List[StreamMetadataError], Dict[str, StreamMetadata]]:
        return self._service.get_program_metadata()

    def __iter__(self) -> Generator[Tuple[str, Optional[BaseStreamingResultFetcher]], None, None]:  # type: ignore[override]
        warnings.warn(
            deprecation_message(
                method="streaming_result_fetcher.__iter__",
                deprecated_in="1.2.0",
                removed_in="1.4.0",
                details="This function is going to change its API to be similar to this of a dictionary, "
                "Use `iterate_results` for the old API.",
            ),
            DeprecationWarning,
            stacklevel=2,
        )
        return self.iterate_results()

    def iterate_results(self) -> Generator[Tuple[str, Optional[BaseStreamingResultFetcher]], None, None]:
        for item in self._schema.items.values():
            yield item.name, self.get(item.name)

    def keys(self) -> KeysView[str]:
        """
        Returns a view of the names of the results
        """
        return self._all_results.keys()

    def items(self) -> ItemsView[str, BaseStreamingResultFetcher]:
        """
        Returns a view, in which the first item is the name of the result and the second is the result
        """
        return self._all_results.items()

    def values(self) -> ValuesView[BaseStreamingResultFetcher]:
        """
        Returns a view of the results
        """
        return self._all_results.values()

    def is_processing(self) -> bool:
        """Check if the job is still processing results

        Returns:
            True if results are still being processed, False otherwise
        """
        key = list(self._all_results.keys())[0]
        return self._all_results[key].is_processing()

    @staticmethod
    def _load_schema(service: Union[JobResultServiceApi, JobResultApi]) -> JobResultSchema:
        response = service.get_job_result_schema()
        return JobResultSchema(
            {
                item.name: JobResultItemSchema(
                    item.name,
                    _parse_dtype(item.simple_d_type),
                    tuple(item.shape),
                    item.is_single,
                    item.expected_count,
                )
                for item in response.items
            }
        )

    def get(
        self, name: str, /, default: Optional[Union[BaseStreamingResultFetcher, _T]] = None
    ) -> Optional[Union[BaseStreamingResultFetcher, _T]]:
        """Get a handle to a named result from [stream_processing][qm.qua._dsl.stream_processing]

        Args:
            name: The named result using in [stream_processing][qm.qua._dsl.stream_processing]
            default: The default value to return if the named result is unknown


        Returns:
            A handle object to the results `MultipleNamedJobResult` or `SingleNamedJobResult` or None if the named results in unknown
        """
        return self._all_results.get(name)

    def __contains__(self, name: object) -> bool:
        if not isinstance(name, str):
            return False
        return name in self._all_results

    def wait_for_all_values(self, timeout: Optional[float] = None) -> bool:
        """Wait until we know all values were processed for all named results

        Args:
            timeout: Timeout for waiting in seconds

        Returns:
            True if all finished successfully, False if any result was closed before done
        """

        def on_iteration() -> bool:
            all_job_states = [result.get_job_state() for result in self._all_results.values()]
            all_done = all(state.done for state in all_job_states)
            any_closed = any(state.closed for state in all_job_states)
            return all_done or any_closed

        def on_complete() -> bool:
            return all(result.get_job_state().done for result in self._all_results.values())

        return run_until_with_timeout(
            on_iteration_callback=on_iteration,
            on_complete_callback=on_complete,
            timeout=timeout if timeout else float("infinity"),
            timeout_message="Job was not done in time",
        )



import logging
import dataclasses
from typing_extensions import runtime_checkable
from typing import Dict, Union, Iterable, Protocol, cast

import betterproto

from qm.grpc.general_messages import MessageLevel

LOG_LEVEL_MAP = {
    MessageLevel.Message_LEVEL_ERROR: logging.ERROR,
    MessageLevel.Message_LEVEL_WARNING: logging.WARN,
    MessageLevel.Message_LEVEL_INFO: logging.INFO,
}


Node = Union[betterproto.Message, Iterable["Node"]]


@runtime_checkable
@dataclasses.dataclass
class DataclassProtocol(Protocol):
    pass


def list_fields(node: Node) -> Dict[str, Node]:
    fields = dataclasses.fields(cast(DataclassProtocol, node))
    output = {}
    for field in fields:
        try:
            field_value = getattr(node, field.name)
        except AttributeError:  # this deals with non-serialized fields in betterproto
            continue
        if isinstance(field_value, Iterable) or (
            isinstance(field_value, betterproto.Message) and betterproto.serialized_on_wire(field_value)
        ):
            output[field.name] = field_value
    return output



from typing import Type

from qm.grpc.v2 import JobServiceStub
from qm.api.v2.base_api_v2 import BaseApiV2
from qm.api.models.server_details import ConnectionDetails


class JobGenericApi(BaseApiV2[JobServiceStub]):
    def __init__(self, connection_details: ConnectionDetails, job_id: str):
        super().__init__(connection_details)
        self._id = job_id

    @property
    def _stub_class(self) -> Type[JobServiceStub]:
        return JobServiceStub


class ElementGenericApi(JobGenericApi):
    def __init__(self, connection_details: ConnectionDetails, job_id: str, element_id: str) -> None:
        super().__init__(connection_details, job_id)
        self._element_id = element_id



import abc
from pathlib import Path
from typing import BinaryIO, Optional


class BinaryAsset(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def for_writing(self) -> BinaryIO:
        raise NotImplementedError()

    @abc.abstractmethod
    def for_reading(self) -> BinaryIO:
        raise NotImplementedError()


class BaseStore(metaclass=abc.ABCMeta):
    """The interface to saving data from a running QM job"""

    def __init__(self) -> None:
        super().__init__()

    @abc.abstractmethod
    def job_named_result(self, job_id: str, name: str) -> BinaryAsset:
        raise NotImplementedError()

    @abc.abstractmethod
    def all_job_results(self, job_id: str) -> BinaryAsset:
        raise NotImplementedError()


class FileBinaryAsset(BinaryAsset):
    def __init__(self, path: Path) -> None:
        super().__init__()
        self._path = path
        self._opened_fd: Optional[BinaryIO] = None

    def close(self) -> None:
        if self._opened_fd:
            self._opened_fd.close()

    def for_writing(self) -> BinaryIO:
        self.close()
        self._opened_fd = self._path.open("wb")
        return self._opened_fd

    def for_reading(self) -> BinaryIO:
        self.close()
        self._opened_fd = self._path.open("rb")
        return self._opened_fd


class SimpleFileStore(BaseStore):
    def __init__(self, root: str = ".") -> None:
        super().__init__()
        self._root = Path(root).absolute()

    def _job_path(self, job_id: str) -> Path:
        path = Path(f"{self._root}/{job_id}")
        path.mkdir(parents=True, exist_ok=True)
        return path

    def job_named_result(self, job_id: str, name: str) -> BinaryAsset:
        return FileBinaryAsset(self._job_path(job_id).joinpath(f"result_{name}.npy"))

    def all_job_results(self, job_id: str) -> BinaryAsset:
        return FileBinaryAsset(self._job_path(job_id).joinpath("results.npz"))



from typing import Union

import betterproto

from qm.exceptions import QmQuaException
from qm.utils.protobuf_utils import Node
from qm.serialization.qua_node_visitor import QuaNodeVisitor
from qm.grpc.qua import (
    QuaProgramRampPulse,
    QuaProgramMeasureProcess,
    QuaProgramBinaryExpression,
    QuaProgramVarRefExpression,
    QuaProgramLiteralExpression,
    QuaProgramAnyScalarExpression,
    QuaProgramSaveStatementSource,
    QuaProgramAnalogMeasureProcess,
    QuaProgramArrayLengthExpression,
    QuaProgramArrayVarRefExpression,
    QuaProgramDigitalMeasureProcess,
    QuaProgramLibFunctionExpression,
    QuaProgramArrayCellRefExpression,
    QuaProgramAnalogTimeDivisionSliced,
    QuaProgramAssignmentStatementTarget,
    QuaProgramAnalogTimeDivisionAccumulated,
    QuaProgramDigitalMeasureProcessCounting,
    QuaProgramLibFunctionExpressionArgument,
    QuaProgramAnalogTimeDivisionMovingWindow,
    QuaProgramBinaryExpressionBinaryOperator,
    QuaProgramAnalogProcessTargetTimeDivision,
    QuaProgramAnalogMeasureProcessRawTimeTagging,
    QuaProgramAnalogMeasureProcessBareIntegration,
    QuaProgramDigitalMeasureProcessRawTimeTagging,
    QuaProgramAnalogMeasureProcessDemodIntegration,
    QuaProgramAnalogMeasureProcessHighResTimeTagging,
    QuaProgramAnalogProcessTargetScalarProcessTarget,
    QuaProgramAnalogProcessTargetVectorProcessTarget,
    QuaProgramAnalogMeasureProcessDualBareIntegration,
    QuaProgramAnalogMeasureProcessDualDemodIntegration,
)


class ExpressionSerializingVisitor(QuaNodeVisitor):
    def __init__(self) -> None:
        self._out = ""
        super().__init__()

    def _default_visit(self, node: Node) -> None:
        type_fullname = f"{type(node).__module__}.{type(node).__name__}"
        print(f"missing expression: {type_fullname}")
        super()._default_visit(node)

    def visit_qm_grpc_qua_QuaProgramLibFunctionExpression(self, node: QuaProgramLibFunctionExpression) -> None:
        args = [ExpressionSerializingVisitor.serialize(arg) for arg in node.arguments]
        if node.library_name == "random":
            self._out = f"call_library_function('{node.library_name}', '{node.function_name}', [{','.join(args)}])"
        else:
            library_name = {
                "util": "Util",
                "math": "Math",
                "cast": "Cast",
            }.get(node.library_name, None)
            function_name = {
                "cond": "cond",
                "unsafe_cast_fixed": "unsafe_cast_fixed",
                "unsafe_cast_bool": "unsafe_cast_bool",
                "unsafe_cast_int": "unsafe_cast_int",
                "to_int": "to_int",
                "to_bool": "to_bool",
                "to_fixed": "to_fixed",
                "mul_fixed_by_int": "mul_fixed_by_int",
                "mul_int_by_fixed": "mul_int_by_fixed",
                "log": "log",
                "pow": "pow",
                "div": "div",
                "exp": "exp",
                "pow2": "pow2",
                "ln": "ln",
                "log2": "log2",
                "log10": "log10",
                "sqrt": "sqrt",
                "inv_sqrt": "inv_sqrt",
                "inv": "inv",
                "MSB": "msb",
                "elu": "elu",
                "aelu": "aelu",
                "selu": "selu",
                "relu": "relu",
                "plrelu": "plrelu",
                "lrelu": "lrelu",
                "sin2pi": "sin2pi",
                "cos2pi": "cos2pi",
                "atan_2pi": "atan_2pi",
                "atan2_2pi": "atan2_2pi",
                "abs": "abs",
                "sin": "sin",
                "cos": "cos",
                "atan": "atan",
                "atan2": "atan2",
                "sum": "sum",
                "max": "max",
                "min": "min",
                "argmax": "argmax",
                "argmin": "argmin",
                "dot": "dot",
            }.get(node.function_name, None)
            if library_name is None:
                raise Exception(f"Unsupported library name {node.library_name}")
            if function_name is None:
                raise Exception(f"Unsupported function name {node.function_name}")

            self._out = f"{library_name}.{function_name}({','.join(args)})"

    def visit_qm_grpc_qua_QuaProgramLibFunctionExpressionArgument(
        self, node: QuaProgramLibFunctionExpressionArgument
    ) -> None:
        name, value = betterproto.which_one_of(node, "argument_oneof")
        if value is not None and name in ("scalar", "array"):
            self._out = ExpressionSerializingVisitor.serialize(value)
        else:
            raise QmQuaException(f"Unknown library function argument {name}")

    def visit_qm_grpc_qua_QuaProgramVarRefExpression(self, node: QuaProgramVarRefExpression) -> None:
        var_ref = betterproto.which_one_of(node, "var_oneof")[1]
        self._out = node.name if isinstance(var_ref, str) else f"IO{var_ref}"

    def visit_qm_grpc_qua_QuaProgramArrayVarRefExpression(self, node: QuaProgramArrayVarRefExpression) -> None:
        self._out = node.name

    def visit_qm_grpc_qua_QuaProgramArrayCellRefExpression(self, node: QuaProgramArrayCellRefExpression) -> None:
        var = ExpressionSerializingVisitor.serialize(node.array_var)
        index = ExpressionSerializingVisitor.serialize(node.index)
        self._out = f"{var}[{index}]"

    def visit_qm_grpc_qua_QuaProgramArrayLengthExpression(self, node: QuaProgramArrayLengthExpression) -> None:
        self._out = f"{node.array.name}.length()"

    def visit_qm_grpc_qua_QuaProgramLiteralExpression(self, node: QuaProgramLiteralExpression) -> None:
        self._out = node.value

    def visit_qm_grpc_qua_QuaProgramAssignmentStatementTarget(self, node: QuaProgramAssignmentStatementTarget) -> None:
        super()._default_visit(node)

    def visit_qm_grpc_qua_QuaProgramRampPulse(self, node: QuaProgramRampPulse) -> None:
        super()._default_visit(node)

    def visit_qm_grpc_qua_QuaProgramMeasureProcess(self, node: QuaProgramMeasureProcess) -> None:
        super()._default_visit(node)

    def visit_qm_grpc_qua_QuaProgramAnalogMeasureProcess(self, node: QuaProgramAnalogMeasureProcess) -> None:
        super()._default_visit(node)

    def visit_qm_grpc_qua_QuaProgramAnalogMeasureProcessDemodIntegration(
        self, node: QuaProgramAnalogMeasureProcessDemodIntegration
    ) -> None:
        name = node.integration.name
        output = node.element_output
        target_name, target_value = betterproto.which_one_of(node.target, "processTarget")

        if isinstance(target_value, QuaProgramAnalogProcessTargetScalarProcessTarget):
            target = ExpressionSerializingVisitor.serialize(target_value)
            self._out = f'demod.full("{name}", {target}, "{output}")'
        elif isinstance(target_value, QuaProgramAnalogProcessTargetVectorProcessTarget):
            target = ExpressionSerializingVisitor.serialize(target_value.array)

            time_name, time_value = betterproto.which_one_of(target_value.time_division, "timeDivision")
            if isinstance(time_value, QuaProgramAnalogTimeDivisionSliced):
                self._out = f'demod.sliced("{name}", {target}, {time_value.samples_per_chunk}, "{output}")'
            elif isinstance(time_value, QuaProgramAnalogTimeDivisionAccumulated):
                self._out = f'demod.accumulated("{name}", {target}, {time_value.samples_per_chunk}, "{output}")'
            elif isinstance(time_value, QuaProgramAnalogTimeDivisionMovingWindow):
                self._out = f'demod.moving_window("{name}", {target}, {time_value.samples_per_chunk}, {time_value.chunks_per_window}, "{output}")'
            else:
                raise Exception(f"Unsupported analog process target {target_name}")
        else:
            raise Exception(f"Unsupported analog process target {target_name}")

    def visit_qm_grpc_qua_QuaProgramAnalogMeasureProcessBareIntegration(
        self, node: QuaProgramAnalogMeasureProcessBareIntegration
    ) -> None:
        name = node.integration.name
        output = node.element_output
        target_name, target_value = betterproto.which_one_of(node.target, "processTarget")

        if isinstance(target_value, QuaProgramAnalogProcessTargetScalarProcessTarget):
            target = ExpressionSerializingVisitor.serialize(target_value)
            self._out = f'integration.full("{name}", {target}, "{output}")'
        elif isinstance(target_value, QuaProgramAnalogProcessTargetVectorProcessTarget):
            target = ExpressionSerializingVisitor.serialize(target_value.array)

            time_name, time_value = betterproto.which_one_of(target_value.time_division, "timeDivision")
            if isinstance(time_value, QuaProgramAnalogTimeDivisionSliced):
                self._out = f'integration.sliced("{name}", {target}, {time_value.samples_per_chunk}, "{output}")'
            elif isinstance(time_value, QuaProgramAnalogTimeDivisionAccumulated):
                self._out = f'integration.accumulated("{name}", {target}, {time_value.samples_per_chunk}, "{output}")'
            elif isinstance(time_value, QuaProgramAnalogTimeDivisionMovingWindow):
                self._out = f'integration.moving_window("{name}", {target}, {time_value.samples_per_chunk}, {time_value.chunks_per_window}, "{output}")'
            else:
                raise Exception(f"Unsupported analog process target {target_name}")
        else:
            raise Exception(f"Unsupported analog process target {target_name}")

    def _common_dual_measurement(
        self,
        node: Union[
            QuaProgramAnalogMeasureProcessDualDemodIntegration, QuaProgramAnalogMeasureProcessDualBareIntegration
        ],
        demod_name: str,
    ) -> None:
        name1 = node.integration1.name
        name2 = node.integration2.name
        output1 = node.element_output1
        output2 = node.element_output2
        target_name, target_value = betterproto.which_one_of(node.target, "processTarget")
        if output1 == "out1" and output2 == "out2":
            common_args = f'"{name1}", "{name2}"'
        elif output1 == "out2" and output2 == "out1":
            common_args = f'"{name2}", "{name1}"'
        else:
            common_args = f'"{name1}", "{output1}", "{name2}", "{output2}"'
        if isinstance(target_value, QuaProgramAnalogProcessTargetScalarProcessTarget):
            target = ExpressionSerializingVisitor.serialize(target_value)
            self._out = f"{demod_name}.full({common_args}, {target})"
        elif isinstance(target_value, QuaProgramAnalogProcessTargetVectorProcessTarget):
            target = ExpressionSerializingVisitor.serialize(target_value.array)

            time_name, time_value = betterproto.which_one_of(target_value.time_division, "timeDivision")
            if isinstance(time_value, QuaProgramAnalogTimeDivisionSliced):
                self._out = f"{demod_name}.sliced({common_args}, {time_value.samples_per_chunk}, {target})"
            elif isinstance(time_value, QuaProgramAnalogTimeDivisionAccumulated):
                self._out = f"{demod_name}.accumulated({common_args}, {time_value.samples_per_chunk}, {target})"
            elif isinstance(time_value, QuaProgramAnalogTimeDivisionMovingWindow):
                self._out = f"{demod_name}.moving_window({common_args}, {time_value.samples_per_chunk}, {time_value.chunks_per_window}, {target})"
            else:
                raise Exception(f"Unsupported analog process target {time_name}")
        else:
            raise Exception(f"Unsupported analog process target {target_name}")

    def visit_qm_grpc_qua_QuaProgramAnalogMeasureProcessDualDemodIntegration(
        self, node: QuaProgramAnalogMeasureProcessDualDemodIntegration
    ) -> None:
        self._common_dual_measurement(node, "dual_demod")

    def visit_qm_grpc_qua_QuaProgramAnalogMeasureProcessDualBareIntegration(
        self, node: QuaProgramAnalogMeasureProcessDualBareIntegration
    ) -> None:
        self._common_dual_measurement(node, "dual_integration")

    def _common_time_tagging(
        self,
        node: Union[
            QuaProgramAnalogMeasureProcessRawTimeTagging,
            QuaProgramAnalogMeasureProcessHighResTimeTagging,
            QuaProgramDigitalMeasureProcessRawTimeTagging,
        ],
        demod_name: str,
    ) -> None:
        target = ExpressionSerializingVisitor.serialize(node.target)
        target_len = ExpressionSerializingVisitor.serialize(node.target_len)
        max_time = node.max_time
        element_output = node.element_output
        self._out = f'time_tagging.{demod_name}({target}, {max_time}, {target_len}, "{element_output}")'

    def visit_qm_grpc_qua_QuaProgramAnalogMeasureProcessRawTimeTagging(
        self, node: QuaProgramAnalogMeasureProcessRawTimeTagging
    ) -> None:
        self._common_time_tagging(node, "analog")

    def visit_qm_grpc_qua_QuaProgramAnalogMeasureProcessHighResTimeTagging(
        self, node: QuaProgramAnalogMeasureProcessHighResTimeTagging
    ) -> None:
        self._common_time_tagging(node, "high_res")

    def visit_qm_grpc_qua_QuaProgramDigitalMeasureProcess(self, node: QuaProgramDigitalMeasureProcess) -> None:
        super()._default_visit(node)

    def visit_qm_grpc_qua_QuaProgramDigitalMeasureProcessCounting(
        self, node: QuaProgramDigitalMeasureProcessCounting
    ) -> None:
        element_outputs = []
        for element_output in node.element_outputs:
            element_outputs.append(f'"{element_output}"')
        element_outputs_str = ",".join(element_outputs)
        target = ExpressionSerializingVisitor.serialize(node.target)
        max_time = node.max_time
        self._out = f"counting.digital({target}, {max_time}, ({element_outputs_str}))"

    def visit_qm_grpc_qua_QuaProgramDigitalMeasureProcessRawTimeTagging(
        self, node: QuaProgramDigitalMeasureProcessRawTimeTagging
    ) -> None:
        self._common_time_tagging(node, "digital")

    def visit_qm_grpc_qua_QuaProgramAnalogProcessTargetScalarProcessTarget(
        self, node: QuaProgramAnalogProcessTargetScalarProcessTarget
    ) -> None:
        super()._default_visit(node)

    def visit_qm_grpc_qua_QuaProgramAnalogProcessTargetTimeDivision(
        self, node: QuaProgramAnalogProcessTargetTimeDivision
    ) -> None:
        super()._default_visit(node)

    def visit_qm_grpc_qua_QuaProgramAnyScalarExpression(self, node: QuaProgramAnyScalarExpression) -> None:
        super()._default_visit(node)

    def visit_qm_grpc_qua_QuaProgramSaveStatementSource(self, node: QuaProgramSaveStatementSource) -> None:
        super()._default_visit(node)

    def visit_qm_grpc_qua_QuaProgramBinaryExpression(self, node: QuaProgramBinaryExpression) -> None:
        left = ExpressionSerializingVisitor.serialize(node.left)
        right = ExpressionSerializingVisitor.serialize(node.right)
        sop = node.op
        mapping = {
            QuaProgramBinaryExpressionBinaryOperator.ADD: "+",
            QuaProgramBinaryExpressionBinaryOperator.SUB: "-",
            QuaProgramBinaryExpressionBinaryOperator.GT: ">",
            QuaProgramBinaryExpressionBinaryOperator.LT: "<",
            QuaProgramBinaryExpressionBinaryOperator.LET: "<=",
            QuaProgramBinaryExpressionBinaryOperator.GET: ">=",
            QuaProgramBinaryExpressionBinaryOperator.EQ: "==",
            QuaProgramBinaryExpressionBinaryOperator.MULT: "*",
            QuaProgramBinaryExpressionBinaryOperator.DIV: "/",
            QuaProgramBinaryExpressionBinaryOperator.OR: "|",
            QuaProgramBinaryExpressionBinaryOperator.AND: "&",
            QuaProgramBinaryExpressionBinaryOperator.XOR: "^",
            QuaProgramBinaryExpressionBinaryOperator.SHL: "<<",
            QuaProgramBinaryExpressionBinaryOperator.SHR: ">>",
        }
        if sop in mapping:
            op = mapping[sop]
        else:
            raise Exception(f"Unsupported operator {sop}")
        self._out = f"({left}{op}{right})"

    @staticmethod
    def serialize(node: Node) -> str:
        visitor = ExpressionSerializingVisitor()
        visitor.visit(node)
        return visitor._out



from typing import TYPE_CHECKING, List, Union, Sequence

from betterproto.lib.google.protobuf import Value, ListValue

from qm.grpc.qua import QuaResultAnalysis

if TYPE_CHECKING:
    from qm.qua._dsl import _ResultSource


class _OutputStream:
    def __init__(self, input_stream: "_ResultSource", operator_array: Sequence[str], tag: str):
        self._input_stream = input_stream
        self._operator_array: Sequence[Union[List[str], str]] = operator_array
        self.tag = tag

    def to_proto(self) -> List[Union[List[str], str]]:
        return list(self._operator_array) + [self._input_stream._to_proto()]


class _ResultAnalysis:
    def __init__(self, result_analysis: QuaResultAnalysis):
        self._result_analysis = result_analysis
        self._saves: List[_OutputStream] = []

    def _add_output_stream(self, tag: str, expression: "_ResultSource", operator_array: List[str]) -> None:
        for save in self._saves:
            if save.tag == tag:
                raise Exception("can not save two streams with the same tag")
        self._saves.append(_OutputStream(expression, operator_array, tag))

    def save(self, tag: str, expression: "_ResultSource") -> None:
        self._add_output_stream(tag, expression, operator_array=["save", tag])

    def save_all(self, tag: str, expression: "_ResultSource") -> None:
        self._add_output_stream(tag, expression, operator_array=["saveAll", tag])

    def auto_save_all(self, tag: str, expression: "_ResultSource") -> None:
        self._add_output_stream(tag, expression, operator_array=["saveAll", tag, "auto"])

    def _add_pipeline(self, output: _OutputStream) -> None:
        proto_output = output.to_proto()
        value = _to_list_value(proto_output)
        self._result_analysis.model.append(value)

    def generate_proto(self) -> None:
        for save in self._saves:
            self._add_pipeline(save)


def _to_list_value(from_list: Sequence[Union[str, List[str]]]) -> ListValue:
    res = ListValue()
    for item in from_list:
        if isinstance(item, str):
            res.values.append(Value(string_value=item))
        elif isinstance(item, list):
            res.values.append(Value(list_value=_to_list_value(item)))
    return res



import random
from functools import wraps
from collections.abc import Iterable

import qm.qua._expressions as _exp
from qm.qua._dsl import assign, declare
from qm.utils import get_iterable_elements_datatype as _get_iterable_elements_datatype


def _library_function(lib_name, func_name):
    def library_decorator(function):
        @wraps(function)
        def wrapper(*args, **kwargs):
            new_args = function(*args, **kwargs)
            return call_library_function(lib_name, func_name, new_args)

        return wrapper

    return library_decorator


def _sanitize_arg(arg):
    if isinstance(arg, Iterable):
        return declare(_get_iterable_elements_datatype(arg), value=arg)
    return arg


def call_library_function(lib_name, func_name, args):
    return _exp.QuaExpression(_exp.lib_func(lib_name, func_name, *[_exp.to_expression(_sanitize_arg(x)) for x in args]))


class Math:
    @staticmethod
    @_library_function("math", "log")
    def log(x, base):
        r"""Computes $\mathrm{log}_{base}(x)$

        Args:
            x: a QUA fixed larger than pow2(-8)=0.00390625
            base: a QUA fixed larger than pow2(1/8)=1.09051

        Returns:
            a QUA fixed
        """
        return (x, base)

    @staticmethod
    @_library_function("math", "pow")
    def pow(base, x):
        r"""Computes ${base}^{x}$.
        Does not support base=1, nor the case where both base=0 & x=0.

        Args:
            base: a non-negative QUA fixed
            x: a QUA fixed

        Returns:
            a QUA fixed
        """
        return (base, x)

    @staticmethod
    @_library_function("math", "div")
    def div(x, y):
        r"""Computes the division between two same-type variables $x/y$

        Args:
            x: a QUA parameter
            y: a QUA parameter not equal to 0

        Returns:
            a QUA fixed
        """
        return (x, y)

    @staticmethod
    @_library_function("math", "exp")
    def exp(x):
        r"""Computes $e^{x}$

        Args:
            x: a QUA fixed smaller than ln(8)=2.0794415416

        Returns:
            a QUA fixed
        """
        return (x,)

    @staticmethod
    @_library_function("math", "pow2")
    def pow2(x):
        r"""Computes $2^{x}$

        Args:
            x: a QUA fixed smaller than 3 (to avoid overflow)

        Returns:
            a QUA fixed
        """
        return (x,)

    @staticmethod
    @_library_function("math", "ln")
    def ln(x):
        r"""Computes $\mathrm{ln}(x)$

        Args:
            x: a QUA fixed larger than exp(-8)=0.0003354627

        Returns:
            a QUA fixed
        """
        return (x,)

    @staticmethod
    @_library_function("math", "log2")
    def log2(x):
        r"""Computes $\mathrm{log}_{2}(x)$

        Args:
            x: a QUA fixed larger than pow2(-8)=0.00390625

        Returns:
            a QUA fixed
        """
        return (x,)

    @staticmethod
    @_library_function("math", "log10")
    def log10(x):
        r"""Computes $\mathrm{log}_{10}(x)$

        Args:
            x: a QUA fixed larger than pow10(-8)=0.00000001

        Returns:
            a QUA fixed
        """
        return (x,)

    @staticmethod
    @_library_function("math", "sqrt")
    def sqrt(x):
        r"""Computes the square root of x

        Args:
            x: a non-negative QUA fixed

        Returns:
            a QUA fixed
        """
        return (x,)

    @staticmethod
    @_library_function("math", "inv_sqrt")
    def inv_sqrt(x):
        r"""Computes the inverse square root of x

        Args:
            x: a QUA fixed larger than 1/64

        Returns:
            a QUA fixed
        """
        return (x,)

    @staticmethod
    @_library_function("math", "inv")
    def inv(x):
        r"""Computes the inverse of x

        Args:
            x: a QUA fixed which is x<=-1/8 or 1/8<x

        Returns:
            a QUA fixed
        """
        return (x,)

    @staticmethod
    @_library_function("math", "MSB")
    def msb(x):
        r"""Finds the index of the most significant bit in the parameter x.
        Notes:

        - Result is independent of sign, for example, +3 and -3 will return the same msb
        - The returned value will be the closet log2, rounded down.

          This is given by $\mathrm{floor}(\mathrm{log}_2(|x|))$.

          For example:

          - msb(0.1) will return -4.
          - msb(5) will return 2.
        - For an integer, msb(0) will return 0.
        - For a fixed point number, msb(0) will return -28.

        Args:
            x: a QUA fixed or a QUA int

        Returns:
            a QUA int
        """
        return (x,)

    @staticmethod
    @_library_function("math", "elu")
    def elu(x):
        r"""Computes the Exponential Linear Unit activation function of x
          $\mathrm{ELU(x)} = \mathrm{max}(0, x) + \mathrm{min}(0, \mathrm{exp}(x)-1)$.

        Args:
            x: a QUA fixed

        Returns:
            a QUA fixed
        """
        return (x,)

    @staticmethod
    @_library_function("math", "aelu")
    def aelu(x):
        r"""Computes faster an approximated Exponential Linear Unit activation function of x
          $\mathrm{aELU}(x) \sim \mathrm{ELU}(x)$

        Args:
            x: a QUA fixed

        Returns:
            a QUA fixed
        """
        return (x,)

    @staticmethod
    @_library_function("math", "selu")
    def selu(x):
        r"""Computes the Scaled Exponential Linear Unit activation function of x
        $\mathrm{SELU}(x) = s*(\mathrm{max}(0, x)+a*\mathrm{min}(0, \mathrm{exp}(x)-1))$,
        $a=1.67326324$, $s=1.05070098$

        Args:
            x: a QUA fixed

        Returns:
            a QUA fixed
        """
        return (x,)

    @staticmethod
    @_library_function("math", "relu")
    def relu(x):
        r"""Computes the Rectified Linear Unit activation function of x
          $\mathrm{ReLU}(x) = \mathrm{max}(0, x)$

        Args:
            x: a QUA fixed

        Returns:
            a QUA fixed
        """
        return (x,)

    @staticmethod
    @_library_function("math", "plrelu")
    def plrelu(x, a):
        r"""Computes the Parametric Leaky Rectified Linear Unit activation function of x
          $\mathrm{PLReLU}(x, a) = \mathrm{max}(0, x)+a*\mathrm{min}(0, x)$

        Args:
            x: a QUA fixed
            a: a QUA fixed

        Returns:
            a QUA fixed
        """
        return (x, a)

    @staticmethod
    @_library_function("math", "lrelu")
    def lrelu(x):
        r"""Computes the Leaky Rectified Linear Unit activation function of x
          $\mathrm{LReLU}(x)=\mathrm{max}(0, x)+0.01*\mathrm{min}(0, x)$

        Args:
            x: a QUA fixed

        Returns:
            a QUA fixed
        """
        return (x,)

    @staticmethod
    @_library_function("math", "sin2pi")
    def sin2pi(x):
        r"""Computes $\mathrm{sin}(2 \pi x)$.
        This is more efficient than `Math.sin(2*np.pi*x)`.
        In addition, this function is immune to overflows: An overflow means that the argument gets a $\pm 16$, which does not change the result due to the periodicity of the sine function.

        Args:
            x (QUA variable of type fixed): the angle in radians

        Returns:
            a QUA fixed
        """
        return (x,)

    @staticmethod
    @_library_function("math", "cos2pi")
    def cos2pi(x):
        r"""Computes $\mathrm{cos}(2 \pi x)$.
        This is more efficient than Math.cos($2 \pi x$).
        In addition, this function is immune to overflows: An overflow means that the argument gets a :math:`\pm 16`, which does not change the result due to the periodcity of the cosine function.

        Args:
            x (QUA variable of type fixed): the angle in radians

        Returns:
            a QUA fixed
        """
        return (x,)

    @staticmethod
    @_library_function("math", "atan_2pi")
    def atan_2pi(x):
        r"""Computes $\mathrm{1/2 \pi * atan}(x)$

        -- Available from QOP 2.4 --

        Args:
            x (a QUA fixed)

        Returns:
            a QUA fixed
        """
        return (x,)

    @staticmethod
    @_library_function("math", "atan2_2pi")
    def atan2_2pi(y, x):
        r"""Computes $\mathrm{1/2 \pi * atan2}(y,x)$

        -- Available from QOP 2.4 --

        Args:
            y (a QUA fixed)
            x (a QUA fixed)

        Returns:
            a QUA fixed
        """
        return (y, x)

    @staticmethod
    @_library_function("math", "abs")
    def abs(x):
        r"""Computes the absolute value of x

        Args:
            x: a QUA variable

        Returns:
            a QUA fixed
        """
        return (x,)

    @staticmethod
    @_library_function("math", "sin")
    def sin(x):
        r"""Computes $\mathrm{sin}(x)$

        Args:
            x (QUA variable of type fixed): the angle in radians

        Returns:
            a QUA fixed
        """
        return (x,)

    @staticmethod
    @_library_function("math", "atan")
    def atan(x):
        r"""Computes $\mathrm{atan}(x)$

        -- Available from QOP 2.4 --

        Args:
            x (a QUA fixed which is -1 <= x <= 1))

        Returns:
            a QUA fixed
        """
        return (x,)

    @staticmethod
    @_library_function("math", "atan2")
    def atan2(y, x):
        r"""Computes $\mathrm{atan2}(y,x)$

        -- Available from QOP 2.4 --

        Args:
            y (a QUA fixed)
            x (a QUA fixed)

        Returns:
            a QUA fixed
        """
        return (y, x)

    @staticmethod
    @_library_function("math", "cos")
    def cos(x):
        r"""Computes $\mathrm{cos}(x)$

        Args:
            x (QUA variable of type fixed): the angle in radians

        Returns:
            a QUA fixed
        """
        return (x,)

    @staticmethod
    @_library_function("math", "sum")
    def sum(x):
        r"""Computes the sum of an array x

        Args:
            x: a QUA array

        Returns:
            the sum of the array, has same type as x
        """
        return (x,)

    @staticmethod
    @_library_function("math", "max")
    def max(x):
        r"""Computes the max of an array x

        Args:
            x: a QUA array

        Returns:
            the max value of the array, has same type as x
        """
        return (x,)

    @staticmethod
    @_library_function("math", "min")
    def min(x):
        r"""Computes the min of an array x

        Args:
            x: a QUA array

        Returns:
            the min value of the array, has same type as x
        """
        return (x,)

    @staticmethod
    @_library_function("math", "argmax")
    def argmax(x):
        r"""Return the index of the maximum of an array

        Args:
            x: a QUA array

        Returns:
            the index of maximum value of array, a QUA Integer
        """
        return (x,)

    @staticmethod
    @_library_function("math", "argmin")
    def argmin(x):
        r"""Return the index of the minimum of an array

        Args:
            x: a QUA array

        Returns:
            the index of minimum value of array, a QUA Integer
        """
        return (x,)

    @staticmethod
    @_library_function("math", "dot")
    def dot(x, y):
        r"""Calculates a dot product of two QUA arrays of identical size.

        Args:
            x: a QUA array
            y: a QUA array

        Returns:
            The dot product of x and y, has same type as x and y

        Example:
            ```python
            assign(c, dot(a, b))
            ```
        """
        return x, y


class Cast:
    @staticmethod
    @_library_function("cast", "mul_int_by_fixed")
    def mul_int_by_fixed(x, y):
        r"""Multiplies an int x by a fixed y, returning an int

        Args:
            x: a QUA integer
            y: a QUA fixed

        Returns:
            a QUA int which equals x*y
        """
        return x, y

    @staticmethod
    @_library_function("cast", "mul_fixed_by_int")
    def mul_fixed_by_int(x, y):
        r"""Multiplies a fixed x by an int y, returning a fixed

        Args:
            x: a QUA fixed
            y: a QUA int

        Returns:
            a QUA fixed which equals x*y
        """
        return x, y

    @staticmethod
    @_library_function("cast", "to_int")
    def to_int(x):
        r"""Casts a variable to int. Supports int, fixed or bool

        Args:
            x: a QUA variable

        Returns:
            a QUA int
        """
        return (x,)

    @staticmethod
    @_library_function("cast", "to_fixed")
    def to_fixed(x):
        r"""Casts a variable to fixed. Supports int, fixed or bool

        Args:
            x: a QUA variable

        Returns:
            a QUA fixed
        """
        return (x,)

    @staticmethod
    @_library_function("cast", "to_bool")
    def to_bool(x):
        r"""Casts a variable to bool. Supports int, fixed or bool

        Args:
            x: a QUA variable

        Returns:
            a QUA bool
        """
        return (x,)

    @staticmethod
    @_library_function("cast", "unsafe_cast_int")
    def unsafe_cast_int(x):
        r"""Treats the given input variable, bitwise, as an integer.
        For a given fixed point number, this is equivalent to multiplying by
        $2^{28}$

        Supports int, fixed or bool.

        Args:
            x: a QUA variable

        Returns:
            a QUA int
        """
        return (x,)

    @staticmethod
    @_library_function("cast", "unsafe_cast_fixed")
    def unsafe_cast_fixed(x):
        r"""Treats the given input variable, bitwise, as a fixed point number.
        For a given integer, this is equivalent to multiplying by $2^{-28}$

        Supports int, fixed or bool.

        Args:
            x: a QUA variable

        Returns:
            a QUA fixed
        """
        return (x,)

    @staticmethod
    @_library_function("cast", "unsafe_cast_bool")
    def unsafe_cast_bool(x):
        r"""Treats the given input variable, bitwise, as a boolean.
        A boolean is determined by the right-most bit, so for a given integer, this is
        equivalent to a parity check.

        Supports int, fixed or bool.

        Warning:
            Saving a boolean number which was unsafely cast from an integer/fixed will give the wrong value in python.

        Args:
            x: a QUA variable

        Returns:
            a QUA bool
        """
        return (x,)


class Util:
    @staticmethod
    @_library_function("util", "cond")
    def cond(condition, true_result, false_result):
        r"""Quick conditional operation. This is equivalent to a ternary operator available in some languages:
        i.e. `a ? b : c`, meaning `b` if `a` is true, or `c` if `a` is false.
        There is less computation overhead (less latency) when running this operation relative to the if conditional.

        Example:
            ```python
            assign(var, cond(a, b, c)) #where a is a boolean expression
            ```
        """
        return condition, true_result, false_result


class Random:
    def __init__(self, seed=None):
        r"""A class for generating pseudo-random numbers in QUA

        Args:
            seed: Optional. An integer seed for the pseudo-random number
                generator.
        """
        self._seed = declare(int, value=seed if seed is not None else random.randrange((1 << 28) - 1))

    def set_seed(self, exp):
        r"""Set the seed for the pseudo-random number generator

        Args:
            exp: a QUA expression
        """
        assign(self._seed, exp)

    @_library_function("random", "rand_int")
    def rand_int(self, max_int):
        r"""Returns a pseudorandom integer in range [0, max_int)

        Args:
            max_int: maximum value

        Example:
            ```python
            a= Random()
            assign(b,a.rand_int(max_int))
            ```
        """
        return self._seed, max_int

    @_library_function("random", "rand_fixed")
    def rand_fixed(self):
        r"""Returns a pseudorandom fixed in range [0.0, 1.0)

        Example:
            ```python
            a= Random()
            assign(b,a.rand_fixed())
            ```
        """
        return (self._seed,)



from typing import List, Tuple, cast

from qm.simulate.interface import SimulatorInterface, SupportedConnectionTypes, get_opx_fem_number
from qm.grpc.frontend import (
    SimulationRequest,
    ExecutionRequestSimulateSimulationInterfaceRawInterface,
    ExecutionRequestSimulateSimulationInterfaceRawInterfaceConnections,
)


class RawInterface(SimulatorInterface[ExecutionRequestSimulateSimulationInterfaceRawInterfaceConnections]):
    """Creates a raw interface for use in [qm.simulate.interface.SimulationConfig][].
    A raw interface defines samples that will be inputted into the OPX inputs.

    Args:
        connections (list):

            List of tuples with the connection. Each tuple should be:

                ``(toController: str, toFEM: int, toPort: int, toSamples: List[float])``
        noisePower: How much noise to add to the input.

    Example:
        ```python
        job = qmm.simulate(config, prog, SimulationConfig(
                          duration=20000,
                          # 500 ns of DC 0.2 V into con1 input 1
                          simulation_interface=RawInterface([("con1", 1, [0.2]*500)])
        ```
    """

    @property
    def connections(self) -> List[Tuple[str, int, int, List[float]]]:
        connections = []
        for connection in self._connections:
            connections.append(
                (connection.from_controller, connection.from_fem, connection.from_port, connection.to_samples)
            )
        return connections

    @classmethod
    def _validate_and_standardize_single_connection(
        cls, connection: SupportedConnectionTypes
    ) -> ExecutionRequestSimulateSimulationInterfaceRawInterfaceConnections:
        if not isinstance(connection, tuple):
            raise Exception("each connection must be of type tuple")
        if len(connection) == 4:
            cls._validate_connection_type(
                connection, [str, int, str, list], "(from_controller, from_fem, from_port, to_samples)"
            )
            tuple_4 = cast(Tuple[str, int, int, List[float]], connection)
            return ExecutionRequestSimulateSimulationInterfaceRawInterfaceConnections(
                from_controller=tuple_4[0], from_fem=tuple_4[1], from_port=tuple_4[2], to_samples=tuple_4[3]
            )
        if len(connection) == 3:
            cls._validate_connection_type(connection, [str, int, list], "(from_controller, from_port, to_samples)")
            tuple_3 = cast(Tuple[str, int, List[float]], connection)
            return ExecutionRequestSimulateSimulationInterfaceRawInterfaceConnections(
                from_controller=tuple_3[0],
                from_fem=get_opx_fem_number(),
                from_port=tuple_3[1],
                to_samples=tuple_3[2],
            )
        raise Exception("connection should be tuple of length of 3 or 4")

    def update_simulate_request(self, request: SimulationRequest) -> SimulationRequest:
        request.simulate.simulation_interface.raw = ExecutionRequestSimulateSimulationInterfaceRawInterface(
            noise_power=self.noisePower, connections=self._connections
        )
        return request



import ssl
from typing import Optional
from dataclasses import field, dataclass

from qm.type_hinting.general import PathLike


@dataclass(frozen=True)
class CredentialOverrides:
    certificate_path: PathLike = field(default="")
    verify_mode: ssl.VerifyMode = field(default=ssl.CERT_REQUIRED)
    check_hostname: bool = field(default=True)


def create_credentials(credentials_override: Optional[CredentialOverrides] = None) -> ssl.SSLContext:
    import certifi

    context = ssl.create_default_context(
        purpose=ssl.Purpose.SERVER_AUTH,
        cafile=certifi.where(),
    )
    context.minimum_version = ssl.TLSVersion.TLSv1_2
    context.set_ciphers("ECDHE+AESGCM:ECDHE+CHACHA20:DHE+AESGCM:DHE+CHACHA20")
    context.set_alpn_protocols(["h2"])
    if ssl.HAS_NPN:
        context.set_npn_protocols(["h2"])

    if credentials_override:
        if credentials_override.certificate_path:
            context.load_verify_locations(credentials_override.certificate_path)
        context.verify_mode = credentials_override.verify_mode
        context.check_hostname = credentials_override.check_hostname

    return context


__all__ = ["create_credentials", "CredentialOverrides"]



import dataclasses
from typing import Optional

from qm.api.models.info import QuaMachineInfo

OPX_FEM_IDX = 1


@dataclasses.dataclass(frozen=True)
class ServerCapabilities:
    has_job_streaming_state: bool
    supports_multiple_inputs_for_element: bool
    supports_analog_delay: bool
    supports_shared_oscillators: bool
    supports_crosstalk: bool
    supports_shared_ports: bool
    supports_input_stream: bool
    supports_new_grpc_structure: bool
    supports_double_frequency: bool
    supports_command_timestamps: bool
    supports_inverted_digital_output: bool
    supports_octave_reset: bool
    supports_sticky_elements: bool
    supports_fast_frame_rotation: bool
    fem_number_in_simulator: int
    supports_api_v2: bool = False
    supports_keeping_dc_offsets: bool = False
    opx1000_fems_return_1_based: bool = True

    @staticmethod
    def build(qua_implementation: Optional[QuaMachineInfo] = None) -> "ServerCapabilities":
        caps = qua_implementation.capabilities if qua_implementation is not None else list()
        if "__qop3" in caps:
            return ServerCapabilities(
                has_job_streaming_state=True,
                supports_multiple_inputs_for_element=True,
                supports_analog_delay=True,
                supports_shared_oscillators=True,
                supports_crosstalk=True,
                supports_shared_ports=True,
                supports_input_stream=True,
                supports_new_grpc_structure=True,
                supports_double_frequency=True,
                supports_command_timestamps=True,
                supports_inverted_digital_output=True,
                supports_sticky_elements=True,
                supports_octave_reset=True,
                supports_fast_frame_rotation=True,
                fem_number_in_simulator=OPX_FEM_IDX,
                supports_api_v2=True,
                opx1000_fems_return_1_based="1_based_fem" in caps,
            )
        return ServerCapabilities(
            has_job_streaming_state="qm.job_streaming_state" in caps,
            supports_multiple_inputs_for_element="qm.multiple_inputs_for_element" in caps,
            supports_analog_delay="qm.analog_delay" in caps,
            supports_shared_oscillators="qm.shared_oscillators" in caps,
            supports_crosstalk="qm.crosstalk" in caps,
            supports_shared_ports="qm.shared_ports" in caps,
            supports_input_stream="qm.input_stream" in caps,
            supports_new_grpc_structure="qm.new_grpc_structure" in caps,
            supports_double_frequency="qm.double_frequency" in caps,
            supports_command_timestamps="qm.play_tag" in caps,
            supports_inverted_digital_output="qm.inverted_digital_output" in caps,
            supports_sticky_elements="qm.sticky_elements" in caps,
            supports_octave_reset="qm.octave_reset" in caps,
            supports_fast_frame_rotation="qm.fast_frame_rotation" in caps,
            fem_number_in_simulator=0,
            supports_keeping_dc_offsets="qm.keep_dc_offsets_when_closing" in caps,
        )



from enum import Enum
from typing import List, Union
from dataclasses import dataclass

from qm.grpc.results_analyser import GetJobErrorsResponseError, GetJobErrorsResponseExecutionErrorSeverity
from qm.grpc.v2 import (
    GetJobErrorsResponseGetJobErrorsResponseSuccessError,
    GetJobErrorsResponseGetJobErrorsResponseSuccessExecutionErrorSeverity,
)


class ExecutionErrorSeverity(Enum):
    Warn = 0
    Error = 1


@dataclass(frozen=True)
class ExecutionError:
    error_code: int
    message: str
    severity: ExecutionErrorSeverity

    def __repr__(self) -> str:
        return f"{self.error_code}\t\t{self.severity.name}\t\t{self.message}"

    @classmethod
    def create_from_grpc_message(
        cls, error: Union[GetJobErrorsResponseError, GetJobErrorsResponseGetJobErrorsResponseSuccessError]
    ) -> "ExecutionError":
        return cls(
            error_code=error.error_code,
            message=error.message,
            severity=cls._parse_severity(error.error_severity),
        )

    @staticmethod
    def _parse_severity(
        error_severity: Union[
            GetJobErrorsResponseExecutionErrorSeverity,
            GetJobErrorsResponseGetJobErrorsResponseSuccessExecutionErrorSeverity,
        ]
    ) -> ExecutionErrorSeverity:
        if error_severity in {
            GetJobErrorsResponseExecutionErrorSeverity.WARNING,
            GetJobErrorsResponseGetJobErrorsResponseSuccessExecutionErrorSeverity.WARNING,
        }:
            return ExecutionErrorSeverity.Warn
        elif error_severity in {
            GetJobErrorsResponseExecutionErrorSeverity.ERROR,
            GetJobErrorsResponseGetJobErrorsResponseSuccessExecutionErrorSeverity.ERROR,
        }:
            return ExecutionErrorSeverity.Error
        raise TypeError(f"No severity level: {error_severity}")


class ExecutionReport:
    def __init__(self, job_id: str, errors: List[ExecutionError]) -> None:
        self._job_id = job_id
        self._errors = errors

    def has_errors(self) -> bool:
        """Returns: True if encountered a runtime error while executing the job."""
        return len(self._errors) > 0

    def errors(self) -> List[ExecutionError]:
        """Returns: list of all execution errors for this job"""
        return self._errors.copy()

    @property
    def _report_header(self) -> str:
        return (
            f"Execution report for job {self._job_id}\nErrors:\n"
            f"Please refer to section: "
            f"Error Indications and Error Reporting in documentation for additional information\n\n"
            "code\t\tseverity\tmessage"
        )

    def __repr__(self) -> str:
        if not self.has_errors():
            return f"Execution report for job {self._job_id}\nNo errors"

        errors_str = self._report_header
        for error in self._errors:
            errors_str += "\n" + str(error)
        return errors_str



from collections import defaultdict
from typing import Dict, List, Union, Literal, Mapping, Optional, Sequence, TypedDict, cast

import numpy as np


def _standardize_port(port: Union[str, int]) -> str:
    # Since we moved to a notation of <fem>-<port>, we want to keep backwards compatibility
    if isinstance(port, int) or "-" not in port:
        port = f"1-{port}"
    return port


class _PortsWaveformsContainer(Dict[str, Sequence[float]]):
    def __init__(self, ports: Mapping[str, Sequence[Union[float, complex]]]):
        data = {}
        for k, v in ports.items():
            standard_port = _standardize_port(k)
            data[standard_port] = np.array(v)  # This is a foul, but we want to keep the same behavior as before.
        super().__init__(data)  # type: ignore[arg-type]

    def __getitem__(self, port: Union[str, int]) -> Sequence[float]:
        return super().__getitem__(_standardize_port(port))

    def __contains__(self, item: object) -> bool:
        if not isinstance(item, (int, str)):
            return False
        return super().__contains__(_standardize_port(item))


class SimulatorControllerSamples:
    def __init__(
        self,
        analog: Mapping[str, Sequence[Union[float, complex]]],
        digital: Mapping[str, Sequence[bool]],
        analog_sampling_rate: Optional[Mapping[str, float]] = None,
    ):
        self.analog = _PortsWaveformsContainer(analog)
        self.digital = _PortsWaveformsContainer(digital)
        self._analog_sampling_rate = (
            analog_sampling_rate if analog_sampling_rate is not None else defaultdict(lambda: 1e9)
        )

    @property
    def analog_sampling_rate(self) -> Mapping[str, float]:
        return self._analog_sampling_rate

    def plot(
        self,
        analog_ports: Optional[Union[str, int, List[Union[str, int]]]] = None,
        digital_ports: Optional[Union[str, int, List[Union[str, int]]]] = None,
    ) -> None:
        """Plots the simulated output of the OPX in the given ports.
        If no ports are given, all active ports are plotted.

        Args:
            analog_ports: Union[None, str, list[str]]
            digital_ports: Union[None, str, list[str]]
        """
        import matplotlib.pyplot as plt

        def calc_t_axis(_samples: Sequence[float], _sampling_rate: float) -> np.typing.NDArray[np.float64]:
            return np.arange(len(_samples)) / _sampling_rate * 1e9

        if isinstance(analog_ports, (str, int)):
            analog_ports = [analog_ports]
        if isinstance(digital_ports, (str, int)):
            digital_ports = [digital_ports]

        analog_to_plot = self.analog.keys() if analog_ports is None else analog_ports
        digital_to_plot = self.digital.keys() if digital_ports is None else digital_ports

        for analog_port in analog_to_plot:
            analog_samples = self.analog[analog_port]
            t_axis = calc_t_axis(analog_samples, self._analog_sampling_rate[str(analog_port)])
            if isinstance(analog_samples[0], complex):
                plt.plot(t_axis, np.real(analog_samples), label=f"Analog {analog_port} I")
                plt.plot(t_axis, np.imag(analog_samples), label=f"Analog {analog_port} Q")
            else:
                plt.plot(t_axis, analog_samples, label=f"Analog {analog_port}")
        for digital_port in digital_to_plot:
            digital_samples = self.digital[digital_port]
            plt.plot(calc_t_axis(digital_samples, 1e9), digital_samples, label=f"Digital {digital_port}")
        plt.xlabel("Time [ns]")
        plt.ylabel("Output")
        plt.legend()


class SimulatorSamplesDictType(TypedDict):
    analog: Dict[str, Sequence[float]]
    digital: Dict[str, Sequence[bool]]


class SimulatorSamples(Dict[str, SimulatorControllerSamples]):
    def __getattr__(self, item: str) -> SimulatorControllerSamples:
        return self[item]

    @property
    def __dict__(self) -> Dict[str, SimulatorControllerSamples]:  # type: ignore[override]
        """
        This property is here for backwards compatibility.
        We didn't put a deprecation message because it's annoying.
        In any case, this object is already a dict, so no need to convert it to a dict.
        """
        return {k: v for k, v in self.items()}

    @staticmethod
    def from_np_array(arr: np.typing.NDArray[np.generic]) -> "SimulatorSamples":
        controllers: Dict[str, SimulatorSamplesDictType] = {}
        assert arr.dtype.names is not None
        for col in arr.dtype.names:
            controller_name, output, key = col.split(":")
            output = cast(Literal["analog", "digital"], output)
            controller = controllers.setdefault(controller_name, {"analog": {}, "digital": {}})
            controller[output][key] = arr[col]  # type: ignore[call-overload]
        res = {}
        for controller_name, samples in controllers.items():
            res[controller_name] = SimulatorControllerSamples(samples["analog"], samples["digital"])
        return SimulatorSamples(res)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: qm/pb/general_messages.proto
# plugin: python-betterproto
# This file has been @generated

from dataclasses import dataclass

import betterproto


class MessageLevel(betterproto.Enum):
    Message_LEVEL_ERROR = 0
    Message_LEVEL_WARNING = 1
    Message_LEVEL_INFO = 2


@dataclass(eq=False, repr=False)
class ErrorMessage(betterproto.Message):
    code: int = betterproto.int32_field(1)
    message: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class Matrix(betterproto.Message):
    v00: float = betterproto.double_field(1)
    v01: float = betterproto.double_field(2)
    v10: float = betterproto.double_field(3)
    v11: float = betterproto.double_field(4)



from typing import TYPE_CHECKING, Optional

import betterproto

from qm.grpc.job_manager import InsertInputStreamResponse, GetElementCorrectionResponse, SetElementCorrectionResponse

if TYPE_CHECKING:
    from qm.grpc.job_manager import (
        Deadline,
        MetadataLike,
        InsertInputStreamRequest,
        GetElementCorrectionRequest,
        SetElementCorrectionRequest,
    )


class DeprecatedJobManagerServiceStub(betterproto.ServiceStub):
    async def get_element_correction(
        self,
        get_element_correction_request: "GetElementCorrectionRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None,
    ) -> "GetElementCorrectionResponse":
        return await self._unary_unary(
            "/qm.grpc.jobManager.JobManagerService/GetElementCorrection",
            get_element_correction_request,
            GetElementCorrectionResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def set_element_correction(
        self,
        set_element_correction_request: "SetElementCorrectionRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None,
    ) -> "SetElementCorrectionResponse":
        return await self._unary_unary(
            "/qm.grpc.jobManager.JobManagerService/SetElementCorrection",
            set_element_correction_request,
            SetElementCorrectionResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def insert_input_stream(
        self,
        insert_input_stream_request: "InsertInputStreamRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None,
    ) -> "InsertInputStreamResponse":
        return await self._unary_unary(
            "/qm.grpc.jobManager.JobManagerService/InsertInputStream",
            insert_input_stream_request,
            InsertInputStreamResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )



import logging
import warnings
from typing import TYPE_CHECKING, Tuple, Optional

from qm.api.v2.job_api import JobApi
from qm.persistence import BaseStore
from qm.program.program import Program
from qm.utils import deprecation_message
from qm.jobs.job_queue_base import QmQueueBase
from qm.api.models.capabilities import ServerCapabilities
from qm.api.models.compiler import CompilerOptionArguments
from qm.api.v2.job_api.job_api import JobApiWithDeprecations

logger = logging.getLogger(__name__)


if TYPE_CHECKING:
    from qm.api.v2.qm_api_old import QmApiWithDeprecations


class QmQueueWithDeprecations(QmQueueBase[JobApi]):
    def __init__(self, store: BaseStore, api: "QmApiWithDeprecations", capabilities: ServerCapabilities):
        super().__init__(store, capabilities)
        self._api = api

    @property
    def pending_jobs(self) -> Tuple[JobApiWithDeprecations, ...]:
        warnings.warn(
            deprecation_message(
                method="queue.pending_jobs",
                deprecated_in="1.2.0",
                removed_in="1.4.0",
                details='This property is going to be removed, use qm.get_jobs("In queue").',
            ),
            DeprecationWarning,
            stacklevel=1,
        )
        return self._get_pending_jobs()

    def _get_pending_jobs(
        self, job_id: Optional[str] = None, position: Optional[int] = None, user_id: Optional[str] = None
    ) -> Tuple[JobApiWithDeprecations, ...]:
        if position is not None:
            logger.warning("Position is not supported in the new API")
        new_jobs = self._api.get_jobs(
            job_ids=[job_id] if job_id else [],
            user_ids=[user_id] if user_id else [],
            status=["In queue"],
        )
        result = tuple(self._api.get_job_by_id(job.id) for job in new_jobs)
        return result

    def add(
        self,
        program: Program,
        compiler_options: Optional[CompilerOptionArguments] = None,
    ) -> JobApi:
        """Adds a QmJob to the queue.
        Programs in the queue will play as soon as possible.

        Args:
            program: A QUA program
            compiler_options: Optional arguments for compilation

        Example:
            ```python
            qm.queue.add(program)  # adds at the end of the queue
            qm.queue.insert(program, position)  # adds at position
            ```
        """
        warnings.warn(
            deprecation_message(
                method="queue.add",
                deprecated_in="1.2.0",
                removed_in="1.4.0",
                details="This method is going to be removed, use qm.add_to_queue.",
            ),
            DeprecationWarning,
            stacklevel=1,
        )
        if compiler_options is None:
            compiler_options = CompilerOptionArguments()

        return self._api.add_to_queue(program)

    def add_compiled(self, program_id: str) -> JobApi:
        """Deprecated - This method is going to be removed, use `qm.add_to_queue()`.

        Adds a compiled QUA program to the end of the queueץ
        Programs in the queue will play as soon as possible.
        For a detailed explanation see
        [Precompile Jobs](../Guides/features.md#precompile-jobs).

        Args:
            program_id: A QUA program ID returned from the compile
                function
        """
        warnings.warn(
            deprecation_message(
                method="queue.add_compiled",
                deprecated_in="1.2.0",
                removed_in="1.4.0",
                details="This method is going to be removed, use `qm.add_to_queue()`.",
            ),
            DeprecationWarning,
            stacklevel=1,
        )
        return self._api.add_to_queue(program_id)

    def remove_by_id(self, job_id: str) -> int:
        """Removes the pending job with a specific job id

        Args:
            job_id: a QMJob id

        Returns:
            The number of jobs removed

        Example:
            ```python
            qm.queue.remove_by_id(job_id)
            ```
        """
        warnings.warn(
            deprecation_message(
                method="queue.remove_by_id",
                deprecated_in="1.2.0",
                removed_in="1.4.0",
                details="This method is going to be removed, use qm.clear_queue(user_ids=[user_id]) or job.cancel().",
            ),
            DeprecationWarning,
            stacklevel=1,
        )
        if job_id is None or job_id == "":
            raise ValueError("job_id can not be empty")
        return len(self._api.clear_queue([job_id]))

    def remove_by_user_id(self, user_id: str) -> int:
        """Removes all pending jobs with a specific user id

        Args:
            user_id: a user id

        Returns:
            The number of jobs removed

        Example:
            ```python
            qm.queue.remove_by_id(job_id)
            ```
        """
        warnings.warn(
            deprecation_message(
                method="queue.remove_by_user_id",
                deprecated_in="1.2.0",
                removed_in="1.4.0",
                details="This method is going to be removed, use qm.clear_queue(user_ids=[user_id]) or job.cancel().",
            ),
            DeprecationWarning,
            stacklevel=1,
        )
        return len(self._api.clear_queue(user_ids=[user_id]))

    def clear(self) -> int:
        """Empties the queue from all pending jobs

        Returns:
            The number of jobs removed
        """
        warnings.warn(
            deprecation_message(
                method="queue.clear",
                deprecated_in="1.2.0",
                removed_in="1.4.0",
                details="This method is going to be removed, use qm.clear_queue.",
            ),
            DeprecationWarning,
            stacklevel=1,
        )
        return len(self._api.clear_queue())
